"""
End2end tests of indications from a wbem server. This tests:
1. creating the subscription in the container
2. Creating the pywbemlistener
3. Receiving indications from the server. It uses only OpenPegasus WBEM server
because OpenPegasus defines a provider in its test environment specifically to
send indications based on a method invocation and we have a container
implementation of OpenPegasus.

"""

from __future__ import absolute_import, print_function

import socket
import time
import sys
import os
import datetime
import logging

from pytest import skip

from .utils.pytest_extensions import skip_if_unsupported_capability

# Note: The wbem_connection fixture uses the es_server fixture, and
# due to the way py.test searches for fixtures, it also must be imported.
# pylint: disable=unused-import,wrong-import-order, relative-beyond-top-level
from .utils.pytest_extensions import wbem_connection  # noqa: F401
from pytest_easy_server import es_server  # noqa: F401
from .utils.pytest_extensions import default_namespace  # noqa: F401
# pylint: enable=unused-import,wrong-import-order, relative-beyond-top-level

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
# pylint: disable=relative-beyond-top-level, disable=redefined-builtin
from ..utils import import_installed

pywbem = import_installed('pywbem')
from pywbem import WBEMServer, WBEMListener, CIMClassName, \
    Error, Uint32, WBEMSubscriptionManager, \
    configure_logger  # noqa: E402
# pylint: enable=relative-beyond-top-level, enable=redefined-buitin
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


def test_indications(wbem_connection):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test indication subscription to a container and reception of indications
    from the container.
    """
    skip_if_unsupported_capability(wbem_connection, 'interop')

    # Get a real IP address for the machine to insert into  the
    # indication_destination_host variable. This must be an address available
    # to both the test machine/vm/etc. and the container in which OpenPegasus
    # is running so OpenPegasus can route indications to the listener.
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    indication_dest_host = s.getsockname()[0]

    # Use the empty string host that allows receving indications addressed to
    # any valid valid destination on the system.
    listener_host = ""
    port = 5000
    log = True
    verbose = True

    # define repeat_loop and indication_count variables to rerun the test with
    # each of these pairs. (repeat_loop, indication_count)
    # FUTURE: Expand this to more indications when we resolve issue #3022,
    # the container not returning all of the indications.  This appears to
    # be much more consistent with python 3.7 and 3.7 PACKAGE_LEVEL=minimum

    # Skipping tests for python 2.7 and 3.7 if issues losing indications
    if sys.version_info[0:2] in ((2, 7), (3, 7)):
        skip("Skipping test_indications for python 2.7, 3.7")
    else:
        test_run_loops = ((1, 1), (1, 2), (1, 10), (1, 100), (1, 1000))

    for params in test_run_loops:
        repeat_loop = params[0]
        indication_count = params[1]

        RunIndicationTest(wbem_connection, indication_dest_host, listener_host,
                          port, indication_count, repeat_loop, log,
                          verbose=verbose)


class RunIndicationTest(object):  # pylint: disable=too-many-instance-attributes
    """
    Runs a test that:
    1. Creates a server
    2. Creates a dynamic listener and starts it.
    3. Creates a filter and subscription
    4. Calls the server to execute a method that creates an indication
    5. waits for indications to be received.
    6. Removes the filter and subscription and stops the listener
    """

    # Definition of the indicationfilter.  This is openpegasus specific and uses
    # a class that generates one indication for each call of a method
    indication_test_class = 'Test_IndicationProviderClass'
    test_class_namespace = 'test/TestProvider'
    test_query = 'SELECT * from %s' % indication_test_class

    # If set, outputs all of the requests (apis and html) and responses to
    # pywbem log file. This is not same log as the listener output.
    output_pywbem_log = False

    def __init__(self, wbem_conn, indication_destination_host,
                 listener_host, http_listener_port, requested_indications,
                 repeat_loop, output_log, verbose):
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-arguments
        """
        Sets up and runs the indication test and removes subscriptions at the
        completion of the test.  There no return from this instance, it
        closes with the test complete or with an exception if the setup
        or test fails.

        Parameters:

            wbem_conn (:class:`WBEMConnection`):
                connection object for the target WBEM server. This must be an
                OpenPegasus server because the indication generator is defined
                as part of that server

            indication_destination_host (:term:`string`):
                host name/ip address containing the listener.  This is
                the address to where indications are sent by the WBEM server.
                If the WBEM server is on a different system or in a container
                this must be an address usable by that system/container.
                Localhost is not a valid host name in that case.

            listener_host (:term:`string` or None):
                host name/ip address that will be used by the listener as
                its bind address. If this parameter contains
                a string that will be passed to WBEMListener as the host
                parameter. That name/IP address is used by the the socket
                services to only accept indications addressed to that
                network interface or IP address (depening on the OS, network,
                etc.)
                If None the listener will be forwarded indications addressed
                to any IP address in the containing computer system.

            http_listener_port (:term:`number`):
                listener port for indications sent with http.

            request_indications (:term:`number`):
                Number of indications requested from server for. The default
                is 1 indication.

            repeat_loop (:term:`number`):
                The number of times the request indications method is to
                be sent to the server.  It is sent to request that the
                server generate the number of indications defined by the
                request_indications parameter and when that number of
                indications has been received

            output_log (:class:`py:bool`):
                Flag to indicate whether the server should generate a log.

            verbose (:class:`py:bool`):
                Flag that when set generates diagnostic displays as the
                test proceedes.
        """

        # Test config parameters
        self.conn = wbem_conn
        self.indication_destination_host = indication_destination_host
        self.listener_host = listener_host
        self.http_listener_port = http_listener_port
        self.requested_indications = requested_indications
        self.repeat_loop = repeat_loop
        self.verbose = verbose
        self.logfile = 'pegasusindicationtest.log'
        if self.verbose:
            print("Listener indication_destination_host IP address {}".
                  format(self.indication_destination_host))

        # Define WBEM server CIMClassName for test indication class
        self.indication_test_class_name = \
            CIMClassName(self.indication_test_class,
                         namespace=self.test_class_namespace)

        # subscription server_id
        self.server_id = None

        # Test status parameters
        self.listener = None
        self.received_indication_count = 0

        # Test indication timers, and sequence number
        self.indication_start_time = None
        self.last_indication_time = None
        self.max_time_between_indications = None
        self.last_seq_num = None

        # Setup test
        if output_log:
            if self.verbose:
                print("Configure pywbem logger to a file.")
            configure_logger('all', log_dest='file', detail_level='all',
                             connection=True)

        # Define the ids of the subscription manager, destination and listener
        sub_mgr_id = 'pywbempegasusIndicationTest'
        self.subscription_destination_id = "dest1"
        self.subscription_filter_id = "filter1"

        if os.path.exists(self.logfile):
            if verbose:
                print("Removing old indication logfile {}".format(self.logfile))
            os.remove(self.logfile)

        logging.basicConfig(filename='pegasusindicationtest.log',
                            level=logging.INFO,
                            format='%(asctime)s %(levelname)s: %(message)s')

        try:
            # Setup WBEMServer instance for the WBEM connection
            server = WBEMServer(self.conn)

            # Create subscription_manager here to be sure we can communicate
            # with server before Creating listener, etc.
            sub_mgr = WBEMSubscriptionManager(
                subscription_manager_id=sub_mgr_id)
            self.server_id = sub_mgr.add_server(server)

            # Remove existing subscriptions, destinations and filters
            # for this server_id since we want to rebuild from
            # scratch. This example uses only owned subscriptions, filters, and
            # destinations.  This only does anything in the odd case where
            # we fail during test and do not clean up at end of test.
            self.remove_existing_subscriptions(sub_mgr, self.server_id)
            if self.verbose:
                print("Removed existing subscriptions")

        except Error as er:  # pylint: disable=invalid-name
            print('Error {} communicating with WBEMServer {}'.format(er,
                                                                     self.conn))
            assert False

        self.create_listener()

        self.create_subscription(self.server_id, sub_mgr)

        # Run the test the number of times defined by repeat_loop
        success = self.repeat_test_loop(self.repeat_loop)

        completion_status = "Success" if success else "Failed"
        print("Indication test completed: {}".format(completion_status))

        # Remove server_id which removes owned subscriptions for this server_id
        sub_mgr.remove_server(self.server_id)

        self.listener.stop()

        assert success

    def create_listener(self):
        """
        Create the listener and listener call back and start the listener
        pylint: disable=global-statement
        """

        # Set the host to empty string if None to indicate no bind address
        # meaning that the listener will receive from any network interface.
        if self.listener_host is None:
            self.listener_host = ""

        self.listener = WBEMListener(self.listener_host,
                                     http_port=self.http_listener_port,
                                     https_port=None)

        # Start connect and start listener.
        self.listener.add_callback(self.consume_indication)

        if self.verbose:
            print("Starting listener\n {}".format(self.listener))
        self.listener.start()

    def create_subscription(self, server_id, sub_mgr):
        """
        Create the indication filter, indication destination and indication
        subscription from the instance variables. Displays the created objects
        if verbose set.
        """

        # Create the destination instance
        indication_dest_url = \
            "http://{0}:{1}".format(self.indication_destination_host,
                                    self.http_listener_port)

        destination = sub_mgr.add_destination(
            server_id, indication_dest_url,
            destination_id=self.subscription_destination_id)

        # Create a dynamic alert indication filter
        filter_ = sub_mgr.add_filter(server_id, self.test_class_namespace,
                                     self.test_query,
                                     query_language="DMTF:CQL",
                                     filter_id=self.subscription_filter_id)

        # Create the subscription
        subscriptions = sub_mgr.add_subscriptions(
            server_id, filter_.path, destination_paths=destination.path)

        if self.verbose:
            print("Subscription instances\n {0}\n{1}\n{2}".
                  format(filter_.tomof(), destination.tomof(),
                         subscriptions[0].tomof()))

    def repeat_test_loop(self, repeat_loop):
        """
        Send indication request to server repeatedly and wait for the
        expected number of indications. Returns only when all repeats completed.
        """
        test_result = False
        while repeat_loop > 0:
            repeat_loop += -1
            self.received_indication_count = 0
            self.indication_start_time = None

            # pylint: disable=too-many-function-args
            self.send_request_for_indications(self.indication_test_class_name,
                                              self.requested_indications)

            test_result = self.receive_expected_indications()

            # Test for success of receive indications. and possibly terminate
            if not test_result:
                print("Test failed. Terminating repeat loop.")
                return False

            # Sleep between each run of the test.
            if repeat_loop > 0:
                time.sleep(self.requested_indications / 150)

        return test_result

    def consume_indication(self, indication, host):
        """
        Consume a single indication. This is a callback and may be on another
        thead. It is called each time an indication is received.
        """
        # Count received indications and save received time.
        if self.indication_start_time is None:
            self.indication_start_time = datetime.datetime.now()

        # time for this indication
        if self.last_indication_time is not None:
            time_diff = datetime.datetime.now() - self.last_indication_time
            if self.max_time_between_indications is None:
                self.max_time_between_indications = time_diff
            elif time_diff > self.max_time_between_indications:
                self.max_time_between_indications = time_diff

        seq_num = indication['SequenceNumber']

        if self.last_seq_num is None:
            self.last_seq_num = seq_num
            self.received_indication_count = 1
        else:
            exp_seq_num = self.last_seq_num + 1
            assert seq_num == exp_seq_num, \
                'Missed {0} indications at seq_num={1}. rcvd seq num={2}, ' \
                'expected={3}'. \
                format((seq_num - 1) - self.last_seq_num,
                       self.last_seq_num, seq_num, exp_seq_num)
            self.received_indication_count += 1

        self.last_seq_num = seq_num
        self.last_indication_time = datetime.datetime.now()

        # pylint: disable=logging-format-interpolation
        self.listener.logger.info(
            "Consumed CIM indication #{}: host={}\n{}".
            format(self.received_indication_count, host,
                   indication.tomof()))

    def receive_expected_indications(self):
        """
        Wait for expected indications to be received.  If there is an
        error in the indications received, try to get more information from
        OpenPegasus.
        Returns True if successful reception, else False
        """
        # Wait for indications to be received.
        success = self.wait_for_indications()

        if not success:
            print("Wait for expected indications failed.\n"
                  "Attempting to display pegasus Indication dest queue.\n")

            insts = self.conn.EnumerateInstances(
                'PG_ListenerDestinationQueue', namespace='root/PG_Internal')
            for inst in insts:
                print('ListenerDestinationName={}:\n   QueueFullDropped={} , '
                      'RetryAttemptsExceeded={}, InQueue={}'
                      .format(inst['ListenerDestinationName'],
                              inst['QueueFullDroppedIndications'],
                              inst['RetryAttemptsExceededIndications'],
                              inst['CurrentIndications']))

                if self.verbose:
                    print("Full instnace of OpenPegasus dest queue:\n{}".
                          format(inst.tomof()))
        return success

    def get_elapsed_time(self):
        """
        Compute elapsed time and indications per second

        returns elapsed time in seconds and count of indications per second
        """
        # If indications received
        if self.received_indication_count != self.requested_indications:
            elapsed_time = \
                self.last_indication_time - self.indication_start_time
            elapsed_time = elapsed_time.total_seconds()
            ind_per_sec = self.received_indication_count / \
                elapsed_time if elapsed_time else 0
        else:
            elapsed_time = 0
            ind_per_sec = 0

        return elapsed_time, ind_per_sec

    def wait_for_indications(self):
        """
        Wait for indications to be received. This waits for the expected number
        of indications to be received to match the request to the server and
        includes indication timers and  stall mechanism that will time out if
        no indication has been received in a reasonable amount of time.

        This method will also catch a terminal control-C and terminate this
        method early.

        Finally it waits for a couple of seconds after the receipt of the
        number of expected indications to see if there are any extra unexpected
        indications, flags that in a message but returns successful completion.
        """
        last_indication_count = 0
        return_flag = False
        stall_ctr = 0

        # Wait loop for reception of all of the indications for indication
        # request to the wbem server.  Try block allows ctrl-C exception
        try:
            return_flag = False

            # Loop to wait for indications and exit if all received or
            # exceeds stall count.
            while True:
                time.sleep(1)

                # Finish processing if all requested indications received
                # Note that this allows for more than expected . There is
                # an error message further down in the code
                if self.received_indication_count >= self.requested_indications:
                    elapsed_time, ind_per_sec = self.get_elapsed_time()

                    print('Rcvd {} indications, time={} sec, {:.2f} per sec'
                          .format(self.received_indication_count,
                                  elapsed_time,
                                  ind_per_sec))
                    return_flag = True
                    break

                # If something received,  reloop, reset stall_ctr
                if last_indication_count != self.received_indication_count:
                    last_indication_count = self.received_indication_count
                    stall_ctr = 0

                    print("Rcvd {} indications.".format(last_indication_count))
                    continue

                # Nothing received in last loop and requested number not rcvd.
                stall_ctr += 1
                if self.last_indication_time:
                    stalled_time = datetime.datetime.now() - \
                        self.last_indication_time
                else:
                    stalled_time = datetime.datetime.now() - \
                        self.indication_start_time

                if stall_ctr > 5:
                    elapsed_time, ind_per_sec = self.get_elapsed_time()

                    print('Nothing received for {} sec: received={} '
                          'time={} sec: {:.2f} per sec. stall ctr={}'.
                          format(stalled_time.total_seconds(),
                                 self.received_indication_count,
                                 elapsed_time,
                                 ind_per_sec,
                                 stall_ctr))
                    break

                print("waiting {:.2f} sec. with no indications ct {}".
                      format(stalled_time.total_seconds(), stall_ctr))

        # Because this can be a long loop, catch CTRL-C
        except KeyboardInterrupt:
            print('Ctrl-C terminating wait loop early')

        # If return_flag True, wait and recheck to be sure no extras
        # received.
        if return_flag:
            time.sleep(1)
            if self.received_indication_count != self.requested_indications:
                print('ERROR. Extra indications received {}, requested {}'.
                      format(self.received_indication_count,
                             self.requested_indications))
            if return_flag:
                assert self.received_indication_count == \
                    self.requested_indications
        return return_flag

    def send_request_for_indications(self, class_name, indication_count):
        """
        Send an invokemethod to the WBEM server to initiate the indication
        output.  This is a pegasus specific invokemethod. Note also that the
        way Pegasus works today, often the response for this request does not
        get returned until well after the indication flow has started because
        it operates on the same thread as the response.

        Returns True if invokemethod accepted by server or False if not
        """
        try:
            # Send method to pegasus server to create  required number of
            # indications. This is a pegasus specific class and method
            send_indication_method = "SendTestIndicationsCount"
            send_count_param = 'indicationSendCount'

            if self.verbose:
                print("Send InvokeMethod to WBEM Server:")
                print("  class_name {0} method {1} for {2} indications".
                      format(class_name, send_indication_method,
                             indication_count))

            result = self.conn.InvokeMethod(
                send_indication_method, class_name,
                [(send_count_param, Uint32(indication_count))])

            if result[0] != 0:
                print('Error: SendTestIndicationCount Method error. '
                      'Nonzero return from InvokeMethod={}'
                      .format(result[0]))
                return False
            return True

        except Error as er:  # pylint: disable=invalid-name
            print('Error: Indication Method exception {}'.format(er))
            return False

    @staticmethod
    def remove_existing_subscriptions(sub_mgr, server_id):
        """
        Remove any existing filters, destinations, and filters for the
        this server_id.
        """
        sub_mgr.remove_subscriptions(
            server_id,
            [inst.path for inst in sub_mgr.get_owned_subscriptions(server_id)])

        for filter_inst in sub_mgr.get_owned_filters(server_id):
            sub_mgr.remove_filter(server_id, filter_inst.path)

        sub_mgr.remove_destinations(
            server_id,
            [inst.path for inst in sub_mgr.get_owned_destinations(server_id)])



"""
Resource consumption tests for mocked EnumerateInstances and
EnumerateInstanceNames operations.
"""

import pytest

from .random_objects import tst_random_conn
from .memory_utils import total_sizeof
from .resource_measurement import ResourceMeasurement


@pytest.mark.parametrize(
    "num_props, num_insts", [
        (1, 1),
        (10, 100),
    ]
)
def test_enuminst_mock(num_props, num_insts):
    """
    Resource consumption test for EnumerateInstances operation against
    a mock environment.
    """

    conn, cls, _ = tst_random_conn(num_insts, num_props)

    print("\nResource consumption test: EnumerateInstances with "
          "num_props={}, num_insts={}:".format(num_props, num_insts))

    with ResourceMeasurement() as rm:
        result = conn.EnumerateInstances(cls.classname)

    result_size = total_sizeof(result)

    print("  Elapsed time: {:>12} ms".
          format("{:.3f}".format(rm.elapsed_time_ms)))
    print("  CPU time:     {:>12} ms".format("{:.3f}".format(rm.cpu_time_ms)))
    print("  Peak usage:   {:>12} B".format(rm.peak_size))
    print("  Result size:  {:>12} B".format(result_size))


@pytest.mark.parametrize(
    "num_props, num_insts", [
        (1, 1),
        (10, 100),
    ]
)
def test_enuminstnames_mock(num_props, num_insts):
    """
    Resource consumption test for EnumerateInstanceNames operation against
    a mock environment.
    """

    conn, cls, _ = tst_random_conn(num_insts, num_props)

    print("\nResource consumption test: EnumerateInstanceNames with "
          "num_props={}, num_insts={}:".format(num_props, num_insts))

    with ResourceMeasurement() as rm:
        result = conn.EnumerateInstanceNames(cls.classname)

    result_size = total_sizeof(result)

    print("  Elapsed time: {:>12} ms".
          format("{:.3f}".format(rm.elapsed_time_ms)))
    print("  CPU time:     {:>12} ms".format("{:.3f}".format(rm.cpu_time_ms)))
    print("  Peak usage:   {:>12} B".format(rm.peak_size))
    print("  Result size:  {:>12} B".format(result_size))



"""
Extensions for unittest module.

The ideas for these extensions are based on functions from the comfychair
module by Martin Pool <mbp@samba.org>.
"""

from __future__ import absolute_import

import re
import os

import six

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem._nocasedict import NocaseDict  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


class RegexpMixin(object):
    """Mixin class for classes derived from unittest.TestCase, providing
    assertion functions for regular expression tests."""

    @staticmethod
    def assert_regexp_contains(text, pattern, msg=None):
        """Assert that a string text *contains* a particular regular expression
        pattern.

        Parameters:
            text         string: string text to be searched
            pattern      string: regular expression
            msg          string: additional message text (normally not needed)

        Raises:
            AssertionError if not matched
          """
        assert re.search(pattern, text), \
            "String text does not contain regexp pattern\n" \
            "    text:    %r\n" \
            "    pattern: %r\n" \
            "%s" % (text, pattern, msg)

    @staticmethod
    def assert_regexp_matches(text, pattern, msg=None):
        """Assert that a string text matches a particular regular expression
        pattern.

        Parameters:
            text         string: string text to be matched
            pattern      string: regular expression
            msg          string: additional message text (normally not needed)

        Raises:
            AssertionError if not matched
          """
        assert re.match(pattern, text), \
            "String text does not match regexp pattern\n" \
            "    text:    %r\n" \
            "    pattern: %r\n" \
            "%s" % (text, pattern, msg)


class FileMixin(object):
    """Mixin class for classes derived from unittest.TestCase, providing
    assertion functions for file related tests."""

    @staticmethod
    def assert_file_exists(filename):
        """Test assert file exists"""
        assert os.path.exists(filename), \
            "File does not exist but should: %s" % filename

    @staticmethod
    def assert_file_not_exists(filename):
        """Test assert file exists but should not"""
        assert not os.path.exists(filename), \
            "File exists but should not: %s" % filename


class CIMObjectMixin(object):
    """
    Mixin class for testing CIM objects.
    """

    @staticmethod
    def assert_CIMInstanceName_attrs(
            obj, classname, keybindings, host=None, namespace=None):
        """
        Verify the attributes of the CIMInstanceName object in `obj` against
        expected values passed as the remaining arguments.
        """
        assert obj.classname == classname
        assert obj.keybindings == keybindings
        assert obj.host == host
        assert obj.namespace == namespace

    @staticmethod
    def assert_CIMProperty_attrs(
            obj, name, value, type_=None, class_origin=None,
            array_size=None, propagated=None, is_array=False,
            reference_class=None, qualifiers=None, embedded_object=None):
        """
        Verify the attributes of the CIMProperty object in `obj` against
        expected values passed as the remaining arguments.
        """
        assert obj.name == name
        assert obj.value == value
        assert obj.type == type_
        assert obj.class_origin == class_origin
        assert obj.array_size == array_size
        assert obj.propagated == propagated
        assert obj.is_array == is_array
        assert obj.reference_class == reference_class
        assert obj.qualifiers == NocaseDict(qualifiers)
        assert obj.embedded_object == embedded_object

    def assert_CIMClass_obj(self, act_class, exp_class):
        """
        Verify that two CIMClass objects are equal.
        """
        assert act_class.classname == exp_class.classname
        context = "class %s" % act_class.classname
        assert act_class.superclass == exp_class.superclass
        self.assert_CIMProperty_dict(
            act_class.properties, exp_class.properties, context)
        self.assert_CIMMethod_dict(
            act_class.methods, exp_class.methods, context)
        self.assert_CIMQualifier_dict(
            act_class.qualifiers, exp_class.qualifiers, context)

    def assert_CIMProperty_dict(self, act_properties, exp_properties, context):
        """
        Verify that two dicts of CIMProperty objects are equal.
        """

        self.assert_dict_keys(act_properties, exp_properties,
                              "%s, property names" % context)
        for propname in act_properties:
            self.assert_CIMProperty_obj(
                act_properties[propname], exp_properties[propname],
                "%s, property %s" % (context, propname))

    def assert_CIMMethod_dict(self, act_methods, exp_methods, context):
        """
        Verify that two dicts of CIMMethod objects are equal.
        """

        self.assert_dict_keys(act_methods, exp_methods,
                              "%s, method names" % context)
        for methname in act_methods:
            self.assert_CIMMethod_obj(
                act_methods[methname], exp_methods[methname],
                "%s, method %s" % (context, methname))

    def assert_CIMParameter_dict(self, act_parameters, exp_parameters, context):
        """
        Verify that two dicts of CIMParameter objects are equal.
        """

        self.assert_dict_keys(act_parameters, exp_parameters,
                              "%s, parameter names" % context)
        for parmname in act_parameters:
            self.assert_CIMParameter_obj(
                act_parameters[parmname], exp_parameters[parmname],
                "%s, parameter %s" % (context, parmname))

    def assert_CIMQualifier_dict(self, act_qualifiers, exp_qualifiers, context):
        """
        Verify that two dicts of CIMQualifier objects are equal.
        """

        self.assert_dict_keys(act_qualifiers, exp_qualifiers,
                              "%s, qualifier names" % context)
        for qualname in act_qualifiers:
            self.assert_CIMQualifier_obj(
                act_qualifiers[qualname], exp_qualifiers[qualname],
                "%s, qualifier %s" % (context, qualname))

    def assert_CIMProperty_obj(self, act_property, exp_property, context):
        """
        Verify that two CIMProperty declaration objects are equal.
        """

        assert act_property.name == exp_property.name, \
            "%s (dict key), name attribute: %s (expected: %s)" % \
            (context, act_property.name, exp_property.name)
        assert act_property.value == exp_property.value, \
            "%s, value attribute: %s (expected: %s)" % \
            (context, act_property.value, exp_property.value)
        assert act_property.type == exp_property.type, \
            "%s, type attribute: %s (expected: %s)" % \
            (context, act_property.type, exp_property.type)
        assert act_property.reference_class == exp_property.reference_class, \
            "%s, reference_class attribute: %s (expected: %s)" % \
            (context,
             act_property.reference_class, exp_property.reference_class)
        assert act_property.embedded_object == exp_property.embedded_object, \
            "%s, embedded_object attribute: %s (expected: %s)" % \
            (context,
             act_property.embedded_object, exp_property.embedded_object)
        assert act_property.is_array == exp_property.is_array, \
            "%s, is_array attribute: %s (expected: %s)" % \
            (context, act_property.is_array, exp_property.is_array)
        assert act_property.array_size == exp_property.array_size, \
            "%s, array_size attribute: %s (expected: %s)" % \
            (context, act_property.array_size, exp_property.array_size)
        assert act_property.class_origin == exp_property.class_origin, \
            "%s, class_origin attribute: %s (expected: %s)" % \
            (context, act_property.class_origin, exp_property.class_origin)
        assert act_property.propagated == exp_property.propagated, \
            "%s, propagated attribute: %s (expected: %s)" % \
            (context, act_property.propagated, exp_property.propagated)
        self.assert_CIMQualifier_dict(
            act_property.qualifiers, exp_property.qualifiers, context)

    def assert_CIMMethod_obj(self, act_method, exp_method, context):
        """
        Verify that two CIMMethod objects are equal.
        """

        assert act_method.name == exp_method.name, \
            "%s (dict key), name attribute: %s (expected: %s)" % \
            (context, act_method.name, exp_method.name)
        assert act_method.return_type == exp_method.return_type, \
            "%s, return_type attribute: %s (expected: %s)" % \
            (context, act_method.return_type, exp_method.return_type)
        assert act_method.class_origin == exp_method.class_origin, \
            "%s, class_origin attribute: %s (expected: %s)" % \
            (context, act_method.class_origin, exp_method.class_origin)
        assert act_method.propagated == exp_method.propagated, \
            "%s, propagated attribute: %s methodmethod (expected: %s)" % \
            (context, act_method.propagated, exp_method.propagated)
        self.assert_CIMParameter_dict(
            act_method.parameters, exp_method.parameters, context)
        self.assert_CIMQualifier_dict(
            act_method.qualifiers, exp_method.qualifiers, context)

    def assert_CIMParameter_obj(self, act_parameter, exp_parameter, context):
        """
        Verify that two CIMParameter objects are equal.
        """

        assert act_parameter.name == exp_parameter.name, \
            "%s (dict key), name attribute: %s (expected: %s)" % \
            (context, act_parameter.name, exp_parameter.name)
        assert act_parameter.type == exp_parameter.type, \
            "%s, type attribute: %s (expected: %s)" % \
            (context, act_parameter.type, exp_parameter.type)
        assert act_parameter.reference_class == exp_parameter.reference_class, \
            "%s, reference_class attribute: %s (expected: %s)" % \
            (context,
             act_parameter.reference_class, exp_parameter.reference_class)
        assert act_parameter.is_array == exp_parameter.is_array, \
            "%s, is_array attribute: %s (expected: %s)" % \
            (context, act_parameter.is_array, exp_parameter.is_array)
        assert act_parameter.array_size == exp_parameter.array_size, \
            "%s, array_size attribute: %s (expected: %s)" % \
            (context, act_parameter.array_size, exp_parameter.array_size)
        self.assert_CIMQualifier_dict(
            act_parameter.qualifiers, exp_parameter.qualifiers, context)

    @staticmethod
    def assert_CIMQualifier_obj(act_qualifier, exp_qualifier, context):
        """
        Verify that two CIMQualifier objects are equal.
        """

        assert act_qualifier.name == exp_qualifier.name, \
            "%s (dict key), name attribute: %s (expected: %s)" % \
            (context, act_qualifier.name, exp_qualifier.name)
        assert act_qualifier.value == exp_qualifier.value, \
            "%s, value attribute: %s (expected: %s)" % \
            (context, act_qualifier.value, exp_qualifier.value)
        assert act_qualifier.type == exp_qualifier.type, \
            "%s, type attribute: %s (expected: %s)" % \
            (context, act_qualifier.type, exp_qualifier.type)
        assert act_qualifier.propagated == exp_qualifier.propagated, \
            "%s, propagated attribute: %s (expected: %s)" % \
            (context, act_qualifier.propagated, exp_qualifier.propagated)
        assert act_qualifier.overridable == exp_qualifier.overridable, \
            "%s, overridable attribute: %s (expected: %s)" % \
            (context, act_qualifier.overridable, exp_qualifier.overridable)
        assert act_qualifier.tosubclass == exp_qualifier.tosubclass, \
            "%s, tosubclass attribute: %s (expected: %s)" % \
            (context, act_qualifier.tosubclass, exp_qualifier.tosubclass)
        assert act_qualifier.toinstance == exp_qualifier.toinstance, \
            "%s, toinstance attribute: %s (expected: %s)" % \
            (context, act_qualifier.toinstance, exp_qualifier.toinstance)
        assert act_qualifier.translatable == exp_qualifier.translatable, \
            "%s, translatable attribute: %s (expected: %s)" % \
            (context, act_qualifier.translatable, exp_qualifier.translatable)

    def assert_dict_keys(self, act_dict, exp_dict, context):
        """
        Verify that two dicts have the same set of keys.
        """
        act_keys = set(act_dict.keys())
        exp_keys = set(exp_dict.keys())
        if act_keys != exp_keys:
            missing_keys = exp_keys - act_keys
            added_keys = act_keys - exp_keys
            msg = "Different sets of %s" % context
            if missing_keys:
                msg += ", missing: %s" % list(missing_keys)
            if added_keys:
                msg += ", added: %s" % list(added_keys)
            self.fail(msg)


def assert_copy(cpy, org):
    """
    Assert that the copied object is a copy of the original object, as follows:

    For mutable types, the copied object must be equal to the original object,
    and they must be different objects.

    For immutable types, the copied object must be equal to the original object,
    but they may be the same object or different objects.

    The determination of mutability in this function is not perfect; it is
    optimized for the types that are used in the tests.
    """
    if isinstance(org, (dict, list)):
        # Mutable
        assert cpy == org
        assert id(cpy) != id(org)
    elif isinstance(org, (tuple, six.string_types, bool, int, type(None))):
        # Immutable
        assert cpy == org
    else:
        raise TypeError("assert_copy() does not support type {}".
                        format(type(org)))



"""
This module provides class NocaseList.
"""

from __future__ import print_function, absolute_import

import sys
import os

__all__ = ['NocaseList']

if sys.version_info[0] == 2:
    # pylint: disable=undefined-variable
    _INTEGER_TYPES = (long, int)  # noqa: F821
else:
    _INTEGER_TYPES = (int,)

# This env var is set when building the docs. It causes the methods
# that are supposed to exist only in a particular Python version, not to be
# removed, so they appear in the docs.
BUILDING_DOCS = os.environ.get('BUILDING_DOCS', False)


class NocaseList(list):
    # pylint: disable=line-too-long
    """
    A case-insensitive and case-preserving list.

    The list is case-insensitive: Whenever items of the list are looked up by
    value or item values are compared, that is done case-insensitively. The
    case-insensitivity is defined by performing the lookup or comparison on the
    result of the :meth:`__casefold__` method on the on list items.
    `None` is allowed as a list value and will not be case folded.

    The list is case-preserving: Whenever the value of list items is returned,
    they have the lexical case that was originally specified when adding
    or updating the item.

    Except for the case-insensitivity of its items, it behaves like, and is
    in fact derived from, the built-in :class:`py:list` class in order to
    facilitate type checks.

    The implementation maintains a second list with the casefolded items of
    the inherited list, and ensures that both lists are in sync.

    The list supports serialization via the Python :mod:`py:pickle` module.
    To save space and time, only the originally cased list is serialized.
    """  # noqa E401
    # pylint: enable=line-too-long

    # Methods not implemented:
    #
    # * __getattribute__(self, name): The method inherited from object is used;
    #   no reason to have a different implementation.
    #
    # * __sizeof__(self): The method inherited from list is used; no reason
    #   to have a different implementation.
    #
    # __repr__(): The method inherited from list is used; no reason
    #   to have a different implementation.
    #
    # __getitem__(): The method inherited from list is used; no reason
    #   to have a different implementation.
    #
    # __iter__(): The method inherited from list is used; no reason
    #   to have a different implementation.

    def __init__(self, iterable=()):
        """
        Initialize the list with the items in the specified iterable.
        """
        super(NocaseList, self).__init__(iterable)

        # The _casefolded_list attribute is a list with the same items as the
        # original (inherited) list, except they are casefolded using the
        # __casefold__() method.

        # The following is an optimization based on the assumption that in
        # many cases, casefolding the input list is more expensive than
        # copying it (plus the overhead to check that).
        if isinstance(iterable, NocaseList):
            try:
                # pylint: disable=protected-access
                self._casefolded_list = iterable._casefolded_list.copy()
            except AttributeError:  # No copy() on Python 2
                self._casefolded_list = self._new_casefolded_list(self)
        else:
            self._casefolded_list = self._new_casefolded_list(self)

    def _new_casefolded_list(self, lst):
        """
        Return a casefolded list from the input list.
        """
        result = []
        for value in lst:
            result.append(self._casefolded_value(value))
        return result

    def _casefolded_value(self, value):
        """
        This method returns the casefolded value and handles the case of value
        being `None`.
        """
        if value is None:
            return None
        return self.__casefold__(value)

    @staticmethod
    def __casefold__(value):
        """
        This method implements the case-insensitive behavior of the class.

        It returns a case-insensitive form of the input value by calling a
        "casefold method" on the value. The input value will not be `None`.

        The casefold method called by this method is :meth:`py2:str.lower` on
        Python 2, and on Python 3 it is :meth:`py:str.casefold`, falling back
        to :meth:`py:bytes.lower` if it does not exist.

        This method can be overridden by users in order to change the
        case-insensitive behavior of the class.
        See :ref:`Overriding the default casefold method` for details.

        Parameters:
          value (str or unicode or bytes): Input value. Will not be `None`.

        Returns:
          str or unicode or bytes: Case-insensitive form of the input value.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        try:
            return value.casefold()
        except AttributeError:
            # Either Python 2, or Python 3 and a byte string key
            return value.lower()

    def __getstate__(self):
        """
        Called when pickling the object, see :meth:`py:object.__getstate__`.

        In order to save space and time, only the list with the originally
        cased items is saved, but not the second list with the casefolded
        items.

        On Python 2, the 'pickle' module does not call :meth:`__setstate__`,
        so this optimzation has only be implemented for Python 3.
        """
        # This copies the state of the inherited list even though it is
        # not visible in self.__dict__.
        state = self.__dict__.copy()
        del state['_casefolded_list']
        return state

    def __setstate__(self, state):
        """
        Called when unpickling the object, see :meth:`py:object.__setstate__`.

        On Python 2, the 'pickle' module does not call this method, so this
        optimzation has only be implemented for Python 3.
        """
        self.__dict__.update(state)
        self._casefolded_list = self._new_casefolded_list(self)

    def __setitem__(self, index, value):
        """
        Update the value of the item at an existing index in the list.

        Invoked using ``ncl[index] = value``.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        super(NocaseList, self).__setitem__(index, value)
        self._casefolded_list[index] = self._casefolded_value(value)

    def __delitem__(self, index):
        """
        Delete an item at an existing index from the list.

        Invoked using ``del ncl[index]``.
        """
        super(NocaseList, self).__delitem__(index)
        del self._casefolded_list[index]

    def __contains__(self, value):
        """
        Return a boolean indicating whether the list contains at least one
        item with the value, by looking it up case-insensitively.

        Invoked using ``value in ncl``.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        return self._casefolded_value(value) in self._casefolded_list

    def __add__(self, other):
        """
        Return a new :class:`NocaseList` object that contains the items from
        the left hand operand (``self``) and the items from the right hand
        operand (``other``).

        The right hand operand (``other``) must be an instance of
        :class:`py:list` (including :class:`NocaseList`) or :class:`py:tuple`.
        The operands are not changed.

        Invoked using e.g. ``ncl + other``

        Raises:
          TypeError: The other iterable is not a list or tuple
        """
        if not isinstance(other, (list, tuple)):
            raise TypeError(
                "Can only concatenate list or tuple (not {t}) to NocaseList".
                format(t=type(other)))
        lst = self.copy()
        lst.extend(other)
        return lst

    def __iadd__(self, other):
        """
        Extend the left hand operand (``self``) by the items from the right
        hand operand (``other``).

        The ``other`` parameter must be an iterable but is otherwise not
        restricted in type. Thus, if it is a string, the characters of the
        string are added as distinct items to the list.

        Invoked using ``ncl += other``.
        """
        # Note: It is unusual that the method has to return self, but it was
        # verified that this is necessary.
        self.extend(other)
        return self

    def __mul__(self, number):
        """
        Return a new :class:`NocaseList` object that contains the items from
        the left hand operand (``self``) as many times as specified by the right
        hand operand (``number``).

        A number <= 0 causes the returned list to be empty.

        The left hand operand (``self``) is not changed.

        Invoked using ``ncl * number``.
        """
        if not isinstance(number, _INTEGER_TYPES):
            raise TypeError(
                "Cannot multiply NocaseList by non-integer of type {t}".
                format(t=type(number)))
        lst = NocaseList()
        for _ in range(0, number):
            lst.extend(self)
        return lst

    def __rmul__(self, number):
        """
        Return a new :class:`NocaseList` object that contains the items from
        the right hand operand (``self``) as many times as specified by the left
        hand operand (``number``).

        A number <= 0 causes the returned list to be empty.

        The right hand operand (``self``) is not changed.

        Invoked using ``number * ncl``.
        """
        lst = self * number  # Delegates to __mul__()
        return lst

    def __imul__(self, number):
        """
        Change the left hand operand (``self``) so that it contains the items
        from the original left hand operand (``self``) as many times as
        specified by the right hand operand (``number``).

        A number <= 0 will empty the left hand operand.

        Invoked using ``ncl *= number``.
        """
        # Note: It is unusual that the method has to return self, but it was
        # verified that this is necessary.
        if not isinstance(number, _INTEGER_TYPES):
            raise TypeError(
                "Cannot multiply NocaseList by non-integer of type {t}".
                format(t=type(number)))
        if number <= 0:
            del self[:]
            del self._casefolded_list[:]
        else:
            self_items = list(self)
            for _ in range(0, number - 1):
                self.extend(self_items)
        return self

    def __reversed__(self):
        """
        Return a shallow copy of the list that has its items reversed in order.

        Invoked using ``reversed(ncl)``.
        """
        lst = self.copy()
        lst.reverse()
        return lst

    def __eq__(self, other):
        """
        Return a boolean indicating whether the list and the other list are
        equal, by comparing corresponding list items case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl == other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        if isinstance(other, NocaseList):
            other = other._casefolded_list  # pylint: disable=protected-access
        else:
            other = self._new_casefolded_list(other)
        return self._casefolded_list == other

    def __ne__(self, other):
        """
        Return a boolean indicating whether the list and the other list are
        not equal, by comparing corresponding list items case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl != other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        if isinstance(other, NocaseList):
            other = other._casefolded_list  # pylint: disable=protected-access
        else:
            other = self._new_casefolded_list(other)
        return self._casefolded_list != other

    def __gt__(self, other):
        """
        Return a boolean indicating whether the list is greater than the other
        list, by comparing corresponding list items case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl > other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        if isinstance(other, NocaseList):
            other = other._casefolded_list  # pylint: disable=protected-access
        else:
            other = self._new_casefolded_list(other)
        return self._casefolded_list > other

    def __lt__(self, other):
        """
        Return a boolean indicating whether the list is less than the other
        list, by comparing corresponding list items case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl < other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        if isinstance(other, NocaseList):
            other = other._casefolded_list  # pylint: disable=protected-access
        else:
            other = self._new_casefolded_list(other)
        return self._casefolded_list < other

    def __ge__(self, other):
        """
        Return a boolean indicating whether the list is greater than or
        equal to the other list, by comparing corresponding list items
        case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl >= other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        if isinstance(other, NocaseList):
            other = other._casefolded_list  # pylint: disable=protected-access
        else:
            other = self._new_casefolded_list(other)
        return self._casefolded_list >= other

    def __le__(self, other):
        """
        Return a boolean indicating whether the list is less than or
        equal to the other list, by comparing corresponding list items
        case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl <= other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        if isinstance(other, NocaseList):
            other = other._casefolded_list  # pylint: disable=protected-access
        else:
            other = self._new_casefolded_list(other)
        return self._casefolded_list <= other

    def count(self, value):
        """
        Return the number of times the specified value occurs in the list,
        comparing the value and the list items case-insensitively.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        return self._casefolded_list.count(self._casefolded_value(value))

    def copy(self):
        """
        Return a shallow copy of the list.

        Note: This method is supported on Python 2 and Python 3, even though
        the built-in list class only supports it on Python 3.
        """
        return NocaseList(self)

    def clear(self):
        """
        Remove all items from the list (and return None).

        Note: This method is supported on Python 2 and Python 3, even though
        the built-in list class only supports it on Python 3.
        """
        try:
            super(NocaseList, self).clear()
            self._casefolded_list.clear()
        except AttributeError:
            del self[:]
            del self._casefolded_list[:]

    def index(self, value, start=0, stop=9223372036854775807):
        """
        Return the index of the first item that is equal to the specified
        value, comparing the value and the list items case-insensitively.

        The search is limited to the index range defined by the specified
        ``start`` and ``stop`` parameters, whereby ``stop`` is the index
        of the first item after the search range.

        Raises:
          AttributeError: The value does not have the casefold method.
          ValueError: No such item is found.
        """
        return self._casefolded_list.index(
            self._casefolded_value(value), start, stop)

    def append(self, value):
        """
        Append the specified value as a new item to the end of the list
        (and return None).

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        super(NocaseList, self).append(value)
        self._casefolded_list.append(self._casefolded_value(value))

    def extend(self, iterable):
        """
        Extend the list by the items in the specified iterable
        (and return None).

        Raises:
          AttributeError: A value in the iterable does not have the casefold
            method.
        """
        super(NocaseList, self).extend(iterable)
        # The following is a circumvention for a behavior of the 'pickle' module
        # that during unpickling may call this method on an object that has
        # been created with __new__() without calling __init__().
        try:
            for value in iterable:
                self._casefolded_list.append(self._casefolded_value(value))
        except AttributeError:
            self._casefolded_list = self._new_casefolded_list(self)

    def insert(self, index, value):
        """
        Insert a new item with specified value before the item at the specified
        index (and return None).

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        super(NocaseList, self).insert(index, value)
        self._casefolded_list.insert(index, self._casefolded_value(value))

    def pop(self, index=-1):
        """
        Return the value of the item at the specified index and also remove it
        from the list.
        """
        self._casefolded_list.pop(index)
        return super(NocaseList, self).pop(index)

    def remove(self, value):
        """
        Remove the first item from the list whose value is equal to the
        specified value (and return None), comparing the value and the list
        items case-insensitively.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        self._casefolded_list.remove(self._casefolded_value(value))
        super(NocaseList, self).remove(value)

    def reverse(self):
        """
        Reverse the items in the list in place (and return None).
        """
        super(NocaseList, self).reverse()
        self._casefolded_list = self._new_casefolded_list(self)

    def sort(self, key=None, reverse=False):
        """
        Sort the items in the list in place (and return None).

        The sort is stable, in that the order of two (case-insensitively) equal
        elements is maintained.

        By default, the list is sorted in ascending order of its casefolded
        item values. If a key function is given, it is applied once to each
        casefolded list item and the list is sorted in ascending or
        descending order of their key function values.

        The ``reverse`` flag can be set to sort in descending order.
        """

        def casefolded_key(value):
            """Key function used for sorting"""
            # The function cannot raise AttributeError due to missing casefold
            # method, because the list items have been verified for that when
            # adding them to the list.
            if key:
                return key(self._casefolded_value(value))
            return self._casefolded_value(value)

        super(NocaseList, self).sort(key=casefolded_key, reverse=reverse)
        self._casefolded_list = self._new_casefolded_list(self)


# Remove methods that should only be present in a particular Python version.
# If the documentation is being built, the methods are not removed in order to
# show them in the documentation.
if sys.version_info[0] == 2 and not BUILDING_DOCS:
    del NocaseList.__setstate__
    del NocaseList.__getstate__



"""
Test for memory_utils module.
"""

from __future__ import absolute_import, print_function

import sys
import struct
from collections import deque
import pytest
import pywbem

from ..utils.pytest_extensions import simplified_test_function
from ...resourcetest import memory_utils
from ...resourcetest.memory_utils import total_sizeof

# Literal form {"blah: 0} faster than dict(blah=0) but same functionality
# pylint: disable=use-dict-literal

# pypy's implementation of sys.getsizeof() raises TypeError
try:
    sys.getsizeof(None)
except TypeError:
    ENABLE_TESTS = False
else:
    ENABLE_TESTS = True

if ENABLE_TESTS:

    class EmptyClass(object):
        # pylint: disable=too-few-public-methods
        "Empty class for determining size of its objects"
        pass

    def size_obj(obj):
        """
        Size of the input object, using the same size calculation function as
        used in memory_utils.total_sizeof().
        """
        return sys.getsizeof(obj)

    SIZE_REF = struct.calcsize('P')  # pylint: disable=no-member

    TESTCASES_TOTAL_SIZEOF = [

        # Testcases for total_sizeof()

        # Each list item is a testcase tuple with these items:
        # * desc: Short testcase description.
        # * kwargs: Keyword arguments for the test function:
        #   * obj: Input object for total_sizeof().
        #   * exp_size: Expected size, i.e. result of total_sizeof().
        # * exp_exc_types: Expected exception type(s), or None.
        # * exp_warn_types: Expected warning type(s), or None.
        # * condition: Boolean condition for testcase to run, or 'pdb' for
        #     debugger

        (
            "Test None",
            dict(
                obj=None,
                exp_size=size_obj(None),
            ),
            None, None, True
        ),
        (
            "Test int",
            dict(
                obj=2020,
                exp_size=size_obj(2020),
            ),
            None, None, True
        ),

        (
            "Test empty string",
            dict(
                obj='',
                exp_size=size_obj(''),
            ),
            None, None, True
        ),
        (
            "Test string with one char",
            dict(
                obj='C',
                exp_size=size_obj('C'),
            ),
            None, None, True
        ),
        (
            "Test string with two chars",
            dict(
                obj='CD',
                exp_size=size_obj('CD'),
            ),
            None, None, True
        ),
        (
            "Test unicode string with two chars",
            dict(
                obj=u'CD',
                exp_size=size_obj(u'CD'),
            ),
            None, None, True
        ),
        (
            "Test binary string with two chars",
            dict(
                obj=b'CD',
                exp_size=size_obj(b'CD'),
            ),
            None, None, True
        ),

        (
            "Test empty tuple",
            dict(
                obj=tuple(),
                exp_size=size_obj(tuple()),
            ),
            None, None, True
        ),
        (
            "Test tuple with one int item",
            dict(
                obj=(2000,),
                exp_size=size_obj((42,)) + size_obj(42),
            ),
            None, None, True
        ),
        (
            "Test tuple with two int items",
            dict(
                obj=(2000, 2001),
                exp_size=size_obj((42, 43)) + 2 * size_obj(42),
            ),
            None, None, True
        ),

        (
            "Test empty list",
            dict(
                obj=[],
                exp_size=size_obj([]),
            ),
            None, None, True
        ),
        (
            "Test list with one int item",
            dict(
                obj=[2000],
                exp_size=size_obj([42]) + size_obj(42),
            ),
            None, None, True
        ),
        (
            "Test list with two int items",
            dict(
                obj=[2000, 2001],
                exp_size=size_obj([42, 43]) + 2 * size_obj(42),
            ),
            None, None, True
        ),

        (
            "Test empty dict",
            dict(
                obj={},
                exp_size=size_obj({}),
            ),
            None, None, True
        ),
        (
            "Test dict with one int/int item",
            dict(
                obj={2000: 2001},
                exp_size=size_obj({42: 43}) + 2 * size_obj(42),
            ),
            None, None, True
        ),
        (
            "Test dict with two int/int items",
            dict(
                obj={2000: 2001, 2002: 2003},
                exp_size=size_obj({42: 43, 44: 45}) + 4 * size_obj(42),
            ),
            None, None, True
        ),

        (
            "Test empty set",
            dict(
                obj=set(),
                exp_size=size_obj(set()),
            ),
            None, None, True
        ),
        (
            "Test empty frozenset",
            dict(
                obj=frozenset(),
                exp_size=size_obj(frozenset()),
            ),
            None, None, True
        ),
        (
            "Test empty deque",
            dict(
                obj=deque(),
                exp_size=size_obj(deque()),
            ),
            None, None, True
        ),

        (
            "Test CIMQualifier object (slotted)",
            dict(
                obj=pywbem.CIMQualifier('Q1', type='string', value=None),
                exp_size=(
                    # Note: None is a shared object. total_sizeof() counts the
                    # first None object, and counts the remaining ones as
                    # references.
                    # Note: This is a slotted object, so there is no __dict__
                    # and the size of the slots are already contained in the
                    # size of the object.
                    size_obj(pywbem.CIMQualifier(
                        'Q1', type='string', value=None)) +
                    size_obj(u'Q1') +  # _name
                    size_obj(u'string') +  # _type
                    size_obj(None) +  # _value (first None object)
                    SIZE_REF +  # _propagated
                    SIZE_REF +  # _overridable
                    SIZE_REF +  # _tosubclass
                    SIZE_REF +  # _toinstance
                    SIZE_REF  # _translatable
                ),
            ),
            None, None, 'debug'
        ),
    ]

    @pytest.mark.parametrize(
        "desc, kwargs, exp_exc_types, exp_warn_types, condition",
        TESTCASES_TOTAL_SIZEOF)
    @simplified_test_function
    def test_total_sizeof(testcase, obj, exp_size):
        """
        Test function for total_sizeof().
        """

        if testcase.condition == 'debug':
            print("\nDebug: Testcase: {}".format(testcase.desc))
            memory_utils.DEBUG = True

        # The code to be tested
        size = total_sizeof(obj)

        if testcase.condition == 'debug':
            memory_utils.DEBUG = False

        # Ensure that exceptions raised in the remainder of this function
        # are not mistaken as expected exceptions
        assert testcase.exp_exc_types is None

        if testcase.condition == 'debug':
            print("Debug: Object size: actual={}, expected={}".
                  format(size, exp_size))

        assert size == exp_size



#
# (C) Copyright 2018 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
Test for the WBEMSubscriptionManager class.
"""
from __future__ import absolute_import, print_function

import os
import re
import pytest

from ...utils import skip_if_moftab_regenerated
from ..utils.dmtf_mof_schema_def import DMTF_TEST_SCHEMA_VER
from ..utils.wbemserver_mock import WbemServerMock
from ..utils.pytest_extensions import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import WBEMServer, CIMClassName, WBEMSubscriptionManager, \
    CIMInstance, CIMDateTime, CIMError, CIM_ERR_ALREADY_EXISTS  # noqa: E402
from pywbem._subscription_manager import SUBSCRIPTION_CLASSNAME, \
    DESTINATION_CLASSNAME, FILTER_CLASSNAME  # noqa: E402
from pywbem_mock.config import OBJECTMANAGERNAME, \
    SYSTEMNAME  # noqa: E402
pywbem_mock = import_installed('pywbem_mock')
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# Literal form {"blah: 0} faster than dict(blah=0) but same functionality
# pylint: disable=use-dict-literal

# Location of DMTF schema directory used by all tests.
# This directory is permanent and should not be removed.
TESTSUITE_SCHEMA_DIR = os.path.join('tests', 'schema')

# The following dictionary is the mock WBEM server configuration for the
# WBEMServerMock class.  It defines a mock WBEM server that is build with
# defined class schemas, classnames, etc.  It uses the same dictionary pattern
# as the default dictionary in the ../utils directory.

SUBSCRIPTION_WBEM_SERVER_MOCK_DICT = {
    'dmtf_schema': {'version': DMTF_TEST_SCHEMA_VER,
                    'dir': TESTSUITE_SCHEMA_DIR},
    'tst_schema': {
        'dir': os.path.join(TESTSUITE_SCHEMA_DIR, 'FakeWBEMServer'),
        'files': []},

    # url of the mock server.  if None the default pywbem_mock url is used
    # This url can be overridden by the WBEMServerMock initialization parameter.
    'url': None,

    'class_names': ['CIM_Namespace',
                    'CIM_ObjectManager',
                    'CIM_RegisteredProfile',
                    'CIM_ElementConformsToProfile',
                    'CIM_ReferencedProfile',
                    'CIM_ComputerSystem',
                    'CIM_IndicationSubscription',
                    'CIM_ListenerDestinationCIMXML',
                    'CIM_IndicationFilter', ],

    'class-mof': ["class XXX_StorageComputerSystem : CIM_ComputerSystem{};", ],
    'system_name': SYSTEMNAME,
    'object_manager': {'Name': OBJECTMANAGERNAME,
                       'ElementName': 'Mock_Test',
                       'Description': 'Mock_Test CIM Server Version 2.15.0'
                                      ' Released', },
    'interop_namspace': 'interop',
    'other_namespaces': [],
    # mock providers that should be installed as part of startup
    # The common parameter is the namespace for the provider
    'providers': ["namespace_provider", 'subscription_providers'],

    # Definition of registered profiles.  Each entry creates a registered
    # profile instance (i.e CIM_RegisteredProfile)
    'registered_profiles': [('DMTF', 'Indications', '1.1.0'),
                            ('DMTF', 'Profile Registration', '1.0.0'),
                            ('SNIA', 'Server', '1.1.0'),
                            ('SNIA', 'Server', '1.2.0'), ],

    'referenced_profiles': [
        (('SNIA', 'Server', '1.2.0'), ('DMTF', 'Indications', '1.1.0')),
    ],
    # Define profile relation to central class instance
    # First element is a profile, second is name and keybindings of a
    # CIMInstanceName
    'element_conforms_to_profile': [],
    # List of CIMInstances. Each entry is a CIM instance with classname,
    # and properties. All properties required to build the path must be
    # defined. No other properties are required for this test.
    'central-instances': [],
    'scoping-instances': []
}


# NOTE: Please add new test functionality to the function
#       test_subscriptionmanager(...)) below. We will leave these functions
#       because they do provide a broad set of test functionality in just a few
#       tests (each tests creates the mock environment)
class BaseMethodsForTests(object):
    """
    Base Class for tests contains all common methods and class level
    setup.
    """

    def setup_method(self):
        """
        Create the Fake connection and setup the connection. This
        method first initializes the WbemServerMock and then supplements
        it with the classes required for the subscription manager.
        """
        skip_if_moftab_regenerated()

        server = WbemServerMock(
            interop_ns='interop',
            server_mock_data=SUBSCRIPTION_WBEM_SERVER_MOCK_DICT)
        # pylint: disable=attribute-defined-outside-init
        self.conn = server.wbem_server.conn

    @staticmethod
    def inst_from_classname(conn, class_name, namespace=None,
                            property_list=None,
                            property_values=None,
                            include_missing_properties=True,
                            include_path=True):
        # TODO AM 8/18 The inst_from_classname() method is not used.
        """
        Build instance from class using class_name property to get class
        from a repository.
        """
        cls = conn.GetClass(class_name, namespace=namespace, LocalOnly=False,
                            IncludeQualifiers=True, include_class_origin=True,
                            property_list=property_list)

        return CIMInstance.from_class(
            cls, namespace=namespace, property_values=property_values,
            include_missing_properties=include_missing_properties,
            include_path=include_path)

    def add_filter(self, sub_mgr, server_id, owned, filter_id=None, name=None):
        """
        Create a single filter definition in the sub_mgr (which adds it to
        the repository) and returns the path of the new filter instance.
        This creates a filter specifically these tests
        """
        # pylint: disable=attribute-defined-outside-init
        self.test_class = 'Test_IndicationProviderClass'
        self.test_class_namespace = 'test/TestProvider'

        self.test_query = 'SELECT * from %s' % self.test_class
        self.test_classname = CIMClassName(self.test_class,
                                           namespace=self.test_class_namespace)

        kwargs = {}
        if filter_id is not None:
            kwargs['filter_id'] = filter_id
        if name is not None:
            kwargs['name'] = name
        filter_ = sub_mgr.add_filter(server_id,
                                     self.test_class_namespace,
                                     self.test_query,
                                     query_language="DMTF:CQL",
                                     owned=owned, **kwargs)
        return filter_.path

    def get_owned_inst_counts(self):
        """
        Using the server connection, get the owned filters, destinations and
        subscriptions from the server and return a tuple with the count for
        each of them.
        """
        server = WBEMServer(self.conn)
        system_name = server.cimom_inst['SystemName']

        owned_dest_pattern = re.compile(r'^pywbemdestination:')
        dest_insts = server.conn.EnumerateInstances(
            DESTINATION_CLASSNAME, namespace=server.interop_ns)
        owned_dest_insts = []
        for inst in dest_insts:
            if re.match(owned_dest_pattern, inst.path.keybindings['Name']) \
                    and inst.path.keybindings['SystemName'] == system_name:
                owned_dest_insts.append(inst)

        filter_insts = server.conn.EnumerateInstances(
            FILTER_CLASSNAME, namespace=server.interop_ns)
        owned_filter_insts = []
        owned_filter_pattern = re.compile(r'^pywbemfilter:')
        for inst in filter_insts:
            if re.match(owned_filter_pattern, inst.path.keybindings['Name']) \
                    and inst.path.keybindings['SystemName'] == system_name:
                owned_filter_insts.append(inst)

        owned_filter_paths = [inst.path for inst in owned_filter_insts]
        owned_dest_paths = [inst.path for inst in owned_dest_insts]

        sub_insts = server.conn.EnumerateInstances(
            SUBSCRIPTION_CLASSNAME, namespace=server.interop_ns)
        owned_sub_insts = []
        for inst in sub_insts:
            if inst.path.keybindings['Filter'] in owned_filter_paths \
                    or inst.path.keybindings['Handler'] in owned_dest_paths:
                owned_sub_insts.append(inst)

        return (len(owned_filter_insts), len(owned_dest_insts),
                len(owned_sub_insts))

    def confirm_created(self, sub_mgr, server_id, filter_path,
                        subscription_paths, owned=True):
        # pylint: disable=no-self-use
        """
        Confirm that filters and subscription paths provided with call
        exist in the appropriated lists.
        """
        if owned:
            owned_filters = sub_mgr.get_owned_filters(server_id)
            owned_filter_paths = [inst.path for inst in owned_filters]

            assert filter_path in owned_filter_paths

            owned_subscriptions = sub_mgr.get_owned_subscriptions(server_id)
            owned_sub_paths = [inst.path for inst in owned_subscriptions]
            for subscription_path in owned_sub_paths:
                assert subscription_path in owned_sub_paths

        all_filters = sub_mgr.get_all_filters(server_id)
        all_filter_paths = [inst.path for inst in all_filters]
        assert filter_path in all_filter_paths

        all_subscriptions = sub_mgr.get_all_subscriptions(server_id)
        all_sub_paths = [inst.path for inst in all_subscriptions]
        for subscription_path in subscription_paths:
            assert subscription_path in all_sub_paths

    def confirm_removed(self, sub_mgr, server_id, filter_paths,
                        subscription_paths):
        # pylint: disable=no-self-use
        """
        When filter_path and subscription path are removed, this
        confirms that results are correct both in the local subscription
        manager and the remote WBEM server.
        """
        if not isinstance(filter_paths, list):
            filter_paths = [filter_paths]

        owned_filters = sub_mgr.get_owned_filters(server_id)
        for filter_path in filter_paths:
            assert filter_path not in owned_filters

        # confirm not in owned subscriptions
        if not isinstance(subscription_paths, list):
            subscription_paths = [subscription_paths]

        owned_subscriptions = sub_mgr.get_owned_subscriptions(server_id)
        owned_sub_paths = [inst.path for inst in owned_subscriptions]
        for subscription_path in subscription_paths:
            assert subscription_path not in owned_sub_paths

        all_filters = sub_mgr.get_all_filters(server_id)
        all_filter_paths = [inst.path for inst in all_filters]
        for filter_path in filter_paths:
            assert filter_path not in all_filter_paths

        all_subscriptions = sub_mgr.get_all_subscriptions(server_id)
        all_sub_paths = [inst.path for inst in all_subscriptions]
        for subscription_path in subscription_paths:
            assert subscription_path not in all_sub_paths

    def empty_expected(self, sub_mgr, server_id):
        """ Get outstanding objects from server. Return True if 0 objects """
        # pylint: disable=no-self-use

        counts = self.get_submgr_inst_counts(sub_mgr, server_id)
        if sum(counts) == 0:
            return True

        print('ERROR: Server_id=%s. Unreleased filters=%s, subs=%s, dest=%s' %
              (server_id, counts[0], counts[1], counts[2]))
        return False

    def get_submgr_inst_counts(self, sub_mgr, server_id):
        """
        Return count of all filters, subscriptions, and destinations on the
        specified subscription manager with the specified server ID,
        and return a tuple with the count for each of them.
        """
        # pylint: disable=no-self-use

        filters = len(sub_mgr.get_all_filters(server_id))
        subscriptions = len(sub_mgr.get_all_subscriptions(server_id))
        dests = len(sub_mgr.get_all_subscriptions(server_id))

        return (filters, subscriptions, dests)


class TestSubMgrClass(BaseMethodsForTests):
    """Test of subscription manager"""

    def test_create_owned_subscription(self):
        """
        Test Basic Creation of sub mgr, creation of owned subscription
        and cleanup based on the WBEMSubscriptionManager context manager.
        """
        sm = "test_create_delete_subscription"
        server = WBEMServer(self.conn)

        listener_url = '%s:%s' % (self.conn.url, 50000)

        with WBEMSubscriptionManager(subscription_manager_id=sm) as sub_mgr:

            server_id = sub_mgr.add_server(server)

            # Create an owned listener, filter, and subscription

            dest = sub_mgr.add_destination(
                server_id, listener_url, owned=True, destination_id='dest1')
            dest_path = dest.path
            # Note: this method returns path.
            filter_path = self.add_filter(
                sub_mgr, server_id, owned=True, filter_id='MyFilterID')
            # Should create a single subscription since there is just one
            # destination
            subscriptions = sub_mgr.add_subscriptions(
                server_id, filter_path, owned=True)

            assert len(subscriptions) == 1
            subscription_paths = [inst.path for inst in subscriptions]

            self.confirm_created(sub_mgr, server_id, filter_path,
                                 subscription_paths, owned=True)

            # Confirm destination instance paths match
            assert len(sub_mgr.get_all_destinations(server_id)) == 1

            # Test that these are considered owned objects
            # Test occurs in confirm_created() but dupped here to confirm
            owned_subscriptions = sub_mgr.get_owned_subscriptions(server_id)
            owned_filters = sub_mgr.get_owned_filters(server_id)
            owned_destinations = sub_mgr.get_owned_destinations(server_id)
            assert dest_path in [inst.path for inst in owned_destinations]
            assert filter_path in [inst.path for inst in owned_filters]
            assert subscription_paths[0] in [inst.path
                                             for inst in owned_subscriptions]

            # Issue #701
            # Test that the same owned subscription is not created a second
            # time because the path is the same as the first.
            subscriptions2 = sub_mgr.add_subscriptions(
                server_id, filter_path, owned=True)
            assert len(subscriptions2) == 1
            # Cannot assert subscription equality because of starttime
            # which is time dependent
            assert [s.path for s in subscriptions] == \
                [s.path for s in subscriptions2]

            # Set times in subscriptions to a fixed time and test for
            # subscription equality
            test_time = CIMDateTime.now()
            for sub in subscriptions:
                sub['SubscriptionStartTime'] = test_time
                sub['TimeOfLastStateChange'] = test_time
            for sub in subscriptions2:
                sub['SubscriptionStartTime'] = test_time
                sub['TimeOfLastStateChange'] = test_time
            assert subscriptions == subscriptions2
            assert len(sub_mgr.get_owned_destinations(server_id)) == 1

            # Trying to create a second filter with the same name fails
            with pytest.raises(CIMError) as exc_info:
                self.add_filter(sub_mgr, server_id, owned=True,
                                filter_id='MyFilterID')
            exc = exc_info.value
            assert exc.status_code == CIM_ERR_ALREADY_EXISTS
            assert len(sub_mgr.get_owned_filters(server_id)) == 1

            sub_mgr.remove_subscriptions(server_id, subscription_paths)
            sub_mgr.remove_filter(server_id, filter_path)

            self.confirm_removed(
                sub_mgr, server_id, [filter_path], subscription_paths)

            assert self.get_submgr_inst_counts(sub_mgr, server_id) == (0, 0, 0)

            sub_mgr.remove_server(server_id)

        # confirm no owned filters, destinations, subscriptions in server
        assert self.get_owned_inst_counts() == (0, 0, 0)

    def test_create_permanent_subscription(self):
        """
        Test creating permanent filter, destination and filter and determining
        if they are retained by the server after the subscription manager
        is closed.
        """
        sm = "test_create_delete_subscription"
        server = WBEMServer(self.conn)

        listener_url = '%s:%s' % (self.conn.url, 50000)
        # Create a single not_owned subscription
        with WBEMSubscriptionManager(subscription_manager_id=sm) as sub_mgr:
            server_id = sub_mgr.add_server(server)
            dest = sub_mgr.add_destination(
                server_id, listener_url, owned=False, name='name1')
            filter_path = self.add_filter(
                sub_mgr, server_id, owned=False, name='MyName')
            subscriptions = sub_mgr.add_subscriptions(
                server_id, filter_path,
                destination_paths=dest.path,
                owned=False)
            subscription_paths = [inst.path for inst in subscriptions]

            self.confirm_created(sub_mgr, server_id, filter_path,
                                 subscription_paths, owned=False)
            assert self.get_submgr_inst_counts(sub_mgr, server_id) == (1, 1, 1)

        # Test that subscriptions instances are still in repo
        # self.conn.display_repository()

        assert self.get_owned_inst_counts() == (0, 0, 0)

        # Create a new submgr and and test for filters, etc. retrieved.
        with WBEMSubscriptionManager(subscription_manager_id=sm) as sub_mgr:
            server_id = sub_mgr.add_server(server)
            assert self.get_submgr_inst_counts(sub_mgr, server_id) == (1, 1, 1)

            sub_paths = [s.path for s in
                         sub_mgr.get_all_subscriptions(server_id)]
            sub_mgr.remove_subscriptions(server_id, sub_paths)

            dest_paths = [d.path for d in
                          sub_mgr.get_all_destinations(server_id)]
            sub_mgr.remove_destinations(
                server_id, dest_paths)

            for filter_ in sub_mgr.get_all_filters(server_id):
                sub_mgr.remove_filter(server_id, filter_.path)
            assert self.get_submgr_inst_counts(sub_mgr, server_id) == (0, 0, 0)

    def test_recovery_of_owned_subscriptions(self):
        """
        Test the ability to recover owned subscriptions from the server. Tests
        ability of SubscriptionManager to recover owned subscriptions, etc.
        from server when started by creating a second subscription mgr.
        """

        sm = "test_subscription_object_recovery"
        server = WBEMServer(self.conn)

        listener_url = '%s:%s' % (self.conn.url, 50000)

        submgr1 = WBEMSubscriptionManager(subscription_manager_id=sm)
        server_id1 = submgr1.add_server(server)

        # Create an owned listener, filter, and subscription

        dest = submgr1.add_destination(
            server_id1, listener_url, owned=True, destination_id='dest1')
        dest_path = dest.path
        filter_path = self.add_filter(
            submgr1, server_id1, owned=True, filter_id='MyFilterID')
        subscriptions = submgr1.add_subscriptions(
            server_id1, filter_path, owned=True)

        assert len(subscriptions) == 1
        subscription_paths = [inst.path for inst in subscriptions]

        self.confirm_created(submgr1, server_id1, filter_path,
                             subscription_paths, owned=True)

        # Confirm destination instance paths match
        assert len(submgr1.get_all_destinations(server_id1)) == 1

        # Test that these are considered owned objects
        # Added to test fix for issue #701,
        # Test already occurs in confirm_created()
        owned_subscriptions = submgr1.get_owned_subscriptions(server_id1)
        owned_filters = submgr1.get_owned_filters(server_id1)
        owned_destinations = submgr1.get_owned_destinations(server_id1)
        assert dest_path in [inst.path for inst in owned_destinations]
        assert filter_path in [inst.path for inst in owned_filters]
        assert subscription_paths[0] in [inst.path
                                         for inst in owned_subscriptions]

        # Create a second subscription manager to recover owned objects
        submgr2 = WBEMSubscriptionManager(subscription_manager_id=sm)
        server_id2 = submgr2.add_server(server)
        assert len(submgr2.get_owned_subscriptions(server_id2)) == 1
        assert len(submgr2.get_owned_destinations(server_id2)) == 1
        assert len(submgr2.get_owned_filters(server_id2)) == 1

        assert submgr2.get_owned_filters(server_id2) == \
            submgr1.get_owned_filters(server_id1)

        assert submgr2.get_owned_destinations(server_id2) == \
            submgr1.get_owned_destinations(server_id1)

        assert submgr2.get_owned_subscriptions(server_id2) == \
            submgr1.get_owned_subscriptions(server_id1)

        # Remove all the subscription objects from submgr1
        submgr1.remove_subscriptions(server_id1, subscription_paths)
        submgr1.remove_filter(server_id1, filter_path)

        self.confirm_removed(submgr1, server_id1, filter_path,
                             subscription_paths)

        assert self.get_submgr_inst_counts(submgr1, server_id1) == (0, 0, 0)

        submgr1.remove_server(server_id1)

        # confirm no owned filters, destinations, subscriptions in server
        assert self.get_owned_inst_counts() == (0, 0, 0)


OK = True
RUN = True
FAIL = False


TESTCASES_SUBMGR = [

    # Testcases for Subscription Tests multiple valid and invalid options
    # in __init__, add_server, add_filter, add_destinations

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * submgr_id: submgr id
    #   * connection_attrs: FakedWBEMConnection attributes. this variable causes
    #     an add_server to be executed. If it is a list, multiple
    #     a server is added for each dict in the list. The attributes of
    #     the FakedWBEMConnection can be set as attributes. Note: this test
    #     already sets the interop namespace automatically
    #   * filter_attrs: dictionary of attributes for add_filter. If it is
    #     a list, one filter is added for each dict in the list
    #   * dest_attrs: dictionary of attributes for add_destination().
    #   * subscription_attrs: dictionary or list dictionaries where each
    #     dictionary is a dictionary that defines either the attributes of
    #     the subscription or to make it simpler to define tests, an integer
    #     for the position in the appropriate list for the required
    #     filter or destination.
    #   * remove_destinations: Option that removes all destinations, if
    #     True, single destination if integer or list of destinations if
    #     list of integers using the saved list from creation with the integer
    #     as index into the list of destinations created.
    #   * remove_filter: Option that removes all filters, if
    #     True, single filter if integer or list of filters if
    #     list of integers using the saved list from creation with the integer
    #     as index into the list of destinations created.
    #   * remove_subscription_attrs: Option that removes all subscriptions, if
    #     True, single subscription if integer or list of filters if
    #     list of integers using the saved list from creation with the integer
    #     as index into the list of subscription created.
    #   * remove_server_attrs: Single server_id identifying server to remove.
    #     This function is executed after all other methods and their tests so
    #     the other tests reflect what was added.
    #   * dest_props: TODO add definition from future pr
    #   * exp_result: Expected dictionary items, for validation. This includes:
    #       * server_id - The expected server_id
    #       * listener_count - Count of expected destinations
    #       * filter_count - Count of expected filters.
    #       * TODO: Add validation for filter, dest attributes
    #       * dest_props - Test for specific properties in specific instances.
    #         TODO: define syntax
    #       * filter_props - Test for specific properties in specific instances.
    #         TODO: define syntax
    #       * subscription_count - Count of expected subscriptions.
    #
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Create no valid id",
        dict(
            submgr_id='MySubscriptionManager',
            connection_attrs=None,
            filter_attrs=None,
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result={}
        ),
        None, None, OK
    ),
    (
        "Invalid subscription manager ID (includes :",
        dict(
            submgr_id='MySubscription:Manager',
            connection_attrs=None,
            filter_attrs=None,
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result={}
        ),
        ValueError, None, OK
    ),
    (
        "Invalid subscription manager id not a string :",
        dict(
            submgr_id=333,
            connection_attrs=None,
            filter_attrs=None,
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result={}
        ),
        TypeError, None, OK
    ),
    (
        "Invalid subscription manager id, value None",
        dict(
            submgr_id=None,
            connection_attrs=None,
            filter_attrs=None,
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result={}
        ),
        ValueError, None, OK
    ),
    (
        "Valid submgr id and server id",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            subscription_attrs=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        None, None, RUN
    ),
    (
        "Valid submgr id and multiple server ids",
        dict(
            submgr_id="ValidID",
            connection_attrs=[dict(url=None),
                              dict(url="http://fake2:5988")],
            filter_attrs=None,
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id=["http://FakedUrl:5988",
                                       "http://fake2:5988"])
        ),
        None, None, OK
    ),
    (
        "Add_filter with No name or filter_id fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id=None, name=None),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add_filter with name and owned fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id=None, name='NotAllowedWithOwned'),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add_filter with filter_id and not owned fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=False,
                              filter_id='id1', name=None),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add_filter with unknown server_id ",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="UnknownServerId",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id=None, name=None),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add_server with both name and filter_id fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Fred", name="fred"),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add_server twice with same serverID",
        dict(
            submgr_id="ValidID",
            connection_attrs=[dict(url=None), dict(url=None)],
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Fred", name="fred"),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add_filter with filter_id containing : character fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Fred:", name=None),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988", )
        ),
        ValueError, None, OK
    ),
    (
        "Add_filter with filter_id not a string fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id=9, name=None),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        TypeError, None, OK
    ),
    (
        "Add_filter with valid filter",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            filter_count=1,
                            filter_props=[0, [('SourceNamespaces',
                                               ['root/interop'])]]),
        ),
        None, None, OK
    ),
    (
        "Add_filter with valid filter that has multiple source namespaces",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces=['root/interop', 'root/cimv2'],
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            filter_count=1,
                            filter_props=[0, [('SourceNamespaces',
                                               ['root/interop',
                                                'root/cimv2'])]]),
        ),
        None, None, OK
    ),
    (
        "Add valid filter with multiple source namespaces, sourcenamespace",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces=['root/interop', 'root/cimv2'],
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None,
                              source_namespace="root/cimv3"),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(
                server_id="http://FakedUrl:5988",
                filter_count=1,
                filter_props=[0, [('SourceNamespaces', ['root/interop',
                                                        'root/cimv2']),
                                  ('SourceNamespace', 'root/cimv3')]], ),
        ),
        None, None, OK
    ),
    (
        "Add_filter with valid filter and remove server_id",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs="http://FakedUrl:5988",
            exp_result=dict(server_id="http://FakedUrl:5988",
                            filter_count=1)
        ),
        None, None, OK
    ),
    (
        "Add_filters with  2 valid filter definitions",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=[dict(server_id="http://FakedUrl:5988",
                               source_namespaces='root/interop',
                               query="SELECT * from blah",
                               query_language='WQL',
                               owned=True,
                               filter_id="Filter1", name=None),
                          dict(server_id="http://FakedUrl:5988",
                               source_namespaces='root/interop',
                               query="SELECT * from blah",
                               query_language='WQL',
                               owned=True,
                               filter_id="Filter2", name=None), ],
            dest_attrs=None,
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            filter_count=2)
        ),
        None, None, OK
    ),
    (
        "Add listener_dest with invalid url",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="httpx",
                            owned=True,
                            destination_id='id1'),
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add listener_dest with invalid persistence_type integer 4",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1',
                            persistence_type=4),
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add duplicate destination. fails because Name property duplicate",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=[dict(server_id="http://FakedUrl:5988",
                             listener_url="http://localhost:5000",
                             owned=True,
                             destination_id='id1',
                             persistence_type='transient'),
                        dict(server_id="http://FakedUrl:5988",
                             listener_url="http://localhost:5000",
                             owned=True,
                             destination_id='id1',
                             persistence_type='transient')],
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        CIMError, None, OK
    ),
    (
        "Add duplicate destination. new Filter_id returns existing instance",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=[dict(server_id="http://FakedUrl:5988",
                             listener_url="http://localhost:5000",
                             owned=True,
                             destination_id='id1',
                             persistence_type='transient'),
                        dict(server_id="http://FakedUrl:5988",
                             listener_url="http://localhost:5000",
                             owned=True,
                             destination_id='id12',
                             persistence_type='transient')],
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            # test only one dest exists
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1, )
        ),
        None, None, OK
    ),
    (
        "Add listener_dest with invalid persistence_type value 'blah",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1',
                            persistence_type='blah'),
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add listener_dest with valid persistence_type string, invalid value",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1',
                            persistence_type="fred"),
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988")
        ),
        ValueError, None, OK
    ),
    (
        "Add listener_dest with valid persistence_type 'transient",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=False,
                            name='name1',
                            persistence_type='transient'),
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            dest_props=[0, [('PersistenceType', 3)]])
        ),
        None, None, OK
    ),
    (
        "Add listener_dest with valid persistence_type 'permanent",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=False,
                            name='name1',
                            persistence_type='permanent'),
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            dest_props=[0, [('PersistenceType', 2)]])
        ),
        None, None, OK
    ),
    (
        "Add listener_dest without persistence_type and owned.",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            dest_props=[0, [('PersistenceType', 3)]])
        ),
        None, None, OK
    ),
    (
        "Add listener_dest without persistence_type owned=False.",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=False,
                            name='name1'),
            subscription_attrs=None,
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            dest_props=[0, [('PersistenceType', 2)]])
        ),
        None, None, OK
    ),
    (
        "Add a single valid subscription",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=True),
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_count=1,
                            subscription_count=1)
        ),
        None, None, OK
    ),
    (
        "Remove listener with existing references",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=True),
            remove_destinations=0,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_count=1,
                            subscription_count=1)
        ),
        CIMError, None, OK
    ),
    (
        "Add a single not_owned subscription with owned filter fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=False),
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_count=1,
                            subscription_count=1)
        ),
        ValueError, None, OK
    ),
    (
        "Add a single not_owned subscription with owned destination fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=False,
                              filter_id="NotOwnedFilter1", name=None),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=False),
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_count=1,
                            subscription_count=1)
        ),
        ValueError, None, OK
    ),
    (
        "Test remove destination succeeds",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            filter_attrs=None,
            subscription_attrs=None,
            remove_destinations=True,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            final_listener_count=0)
        ),
        None, None, OK
    ),
    (
        "Test remove destination with subscriptions fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=True),
            remove_destinations=True,
            remove_filters=None,
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_count=1,
                            subscription_count=1)
        ),
        CIMError, None, OK
    ),
    (
        "Test remove filter succeeds without subscription",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=None,
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            subscription_attrs=None,
            remove_destinations=True,
            remove_filters=[0],  # Removing this filter fails
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            filter_count=1,
                            final_filtercount=1)
        ),
        None, None, OK
    ),
    (
        "Test remove filter fails with subscription",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=True),
            remove_destinations=True,
            remove_filters=[0],  # Removing this filter fails
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_count=1,
                            subscription_count=1)
        ),
        CIMError, None, OK
    ),
    (
        "Test remove destination with subscription fails",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=True),
            remove_destinations=True,
            remove_filters=[0],  # Removing this filter fails
            remove_subscriptions=None,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_count=1,
                            subscription_count=1)
        ),
        CIMError, None, OK
    ),
    (
        "Test remove subscriptions succeeds",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=True),
            remove_destinations=None,
            remove_filters=None,
            remove_subscriptions=True,
            remove_server_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_count=1,
                            subscription_count=1,
                            final_subscription_count=0,
                            final_filter_count=1,
                            final_listener_count=1)
        ),
        None, None, OK
    ),

    # Tests that need to be added
    # Add server where param is not WBEMServer
    # add not_owned filters and destinations
    # NOTE: Eventually this should replace most of the unit tests.
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_SUBMGR)
@simplified_test_function
def test_subscriptionmanager(testcase, submgr_id, connection_attrs,
                             filter_attrs, dest_attrs, subscription_attrs,
                             remove_destinations, remove_filters,
                             remove_subscriptions,
                             remove_server_attrs, exp_result):
    # pylint: disable=unused-argument
    """
    Tests all of the methods of the subscription manager.
    """
    skip_if_moftab_regenerated()

    def define_connection(connection_attrs):
        """
        Create a new mock connection with the attributes defined in
        connection_attrs. and the subscription providers.
        This method also does an add_server with the created mock server.
        Returns:
            returns the server_id associated with the connection.
        """
        mock_server = WbemServerMock(
            interop_ns='interop',
            server_mock_data=SUBSCRIPTION_WBEM_SERVER_MOCK_DICT,
            **connection_attrs)
        return submgr.add_server(mock_server.wbem_server)

    # The code to be tested starts here but encompasses most of this function
    # depending on the function parameters  We did not set a specific place
    # above which there should never be an exception. Depending on the tests,
    # exceptions could occur throughout the code below.
    submgr = WBEMSubscriptionManager(submgr_id)

    # Test result arrays for one or more server_ids. Each of the add and
    # remove methods will insert the results into these lists for validation
    # of the returns from these methods.
    server_id = None
    server_ids = []
    added_filters = []
    added_destinations = []
    added_subscriptions = []

    # The following code adds connections and filters
    # If connection_attrs exists, set up the server and save the server_ids
    # for later use in the test
    if connection_attrs:
        if isinstance(connection_attrs, list):
            for connection_attr in connection_attrs:
                server_ids.append(define_connection(connection_attr))
        else:
            server_id = define_connection(connection_attrs)

    # Test for add filter, adds one or more filters
    if filter_attrs:
        if isinstance(filter_attrs, dict):
            filter_attrs = [filter_attrs]
        for attr in filter_attrs:
            added_filters.append(submgr.add_filter(**attr))

    # Test for add destinations. Adds one or more destinations
    if dest_attrs:
        if isinstance(dest_attrs, dict):
            dest_attrs = [dest_attrs]
        for attr in dest_attrs:
            added_destinations.append(submgr.add_destination(**attr))

    # Test for adding subscriptions based on the attributes in the dest
    # definition. Here, the filter and dest components are integers defining
    # the entry in the already created object lists to use for the path
    # in the subscription.
    if subscription_attrs:
        if isinstance(subscription_attrs, dict):
            subscription_attrs = [subscription_attrs]
        for attr in subscription_attrs:
            if isinstance(attr['filter_path'], int):
                attr['filter_path'] = added_filters[attr['filter_path']].path
            if isinstance(attr['destination_paths'], int):
                attr['destination_paths'] = \
                    added_destinations[attr['destination_paths']].path
            elif isinstance(attr['destination_paths'], list):
                dest_paths = []
                for item in attr['destination_paths']:
                    dest_paths.append(
                        added_destinations[item].path)
                attr['destination_paths'] = dest_paths
            rslts = submgr.add_subscriptions(**attr)
            added_subscriptions.extend(rslts)

    if "server_id" in exp_result:
        if isinstance(exp_result['server_id'], list):
            assert set(server_ids) == set(exp_result['server_id'])
        else:
            assert server_id == exp_result['server_id']

    # test instances created before any remove requests.
    if 'listener_count' in exp_result:
        all_dests = submgr.get_all_destinations(server_id)
        assert len(all_dests) == exp_result['listener_count']

        owned_dests = submgr.get_owned_destinations(server_id)
        dest_prefix = 'pywbemdestination:{0}:'.format(submgr_id,)
        owned = [d for d in all_dests if d['Name'].startswith(dest_prefix)]
        assert set(owned) == set(owned_dests)

    if 'filter_count' in exp_result:
        all_filters = submgr.get_all_filters(server_id)
        assert len(all_filters) == exp_result['filter_count']

    if 'subscription_count' in exp_result:
        all_subs = submgr.get_all_subscriptions(server_id)
        assert len(all_subs) == exp_result['subscription_count']

    # Test for properties in instances
    if 'dest_props' in exp_result:
        dest_props = exp_result['dest_props']
        created_instance = added_destinations[dest_props[0]]
        props = dest_props[1]
        for prop, value in props:
            # If None expect either property with None or no property
            if value is None:
                if prop in created_instance:
                    assert created_instance[prop] == value
            else:
                assert created_instance[prop] == value

    if 'filter_props' in exp_result:
        filter_props = exp_result['filter_props']
        created_instance = added_filters[filter_props[0]]
        props = filter_props[1]
        for prop, value in props:
            # If None expect either property with None or no property
            if value is None:
                if prop in created_instance:
                    assert created_instance[prop] == value
            else:
                if created_instance[prop] != value:
                    print("THEERROR\n{}".format(created_instance.tomof()))
                assert created_instance[prop] == value

    # The removal methods are executed after all others methods defined for
    # the test and after tests on the results of other method execution
    if remove_destinations is not None:
        dest_paths = [inst.path for inst in added_destinations]
        if isinstance(remove_destinations, bool):
            assert remove_destinations
            submgr.remove_destinations(server_id, dest_paths)
            result = 0
        elif isinstance(remove_destinations, int):
            submgr.remove_destinations(server_id,
                                       dest_paths[remove_destinations])
            result = len(added_destinations) - 1
        elif isinstance(remove_destinations, list):
            removal_paths = [added_destinations[i].path for i in
                             remove_destinations]
            submgr.remove_destinations(server_id, removal_paths)
            result = len(added_destinations) - len(remove_destinations)
        assert len(submgr.get_all_destinations(server_id)) == result

    if remove_filters is not None:
        filter_paths = [inst.path for inst in added_filters]
        if isinstance(remove_filters, bool):
            assert remove_filters
            submgr.remove_filter(server_id, filter_paths)
            result = 0
        elif isinstance(remove_filters, int):
            submgr.remove_filter(server_id, filter_paths[remove_filters])
            result = len(added_filters) - 1
        elif isinstance(remove_filters, list):
            # Must remove one by one
            for item in remove_filters:
                submgr.remove_filter(server_id, filter_paths[item])
            result = len(added_filters) - len(remove_filters)
        assert len(submgr.get_all_filters(server_id)) == result

    if remove_subscriptions is not None:
        sub_paths = [inst.path for inst in added_subscriptions]
        if isinstance(remove_subscriptions, bool):
            assert remove_subscriptions
            submgr.remove_subscriptions(server_id, sub_paths)
            result = 0
        elif isinstance(remove_subscriptions, int):
            submgr.remove_subscriptions(server_id,
                                        sub_paths[remove_destinations])
            result = len(sub_paths) - 1
        elif isinstance(remove_subscriptions, list):
            removal_list = [added_subscriptions[i].path for i in
                            remove_subscriptions]
            submgr.remove_subscriptions(server_id, removal_list)
            result = len(sub_paths) - len(remove_subscriptions)
        assert len(submgr.get_all_subscriptions(server_id)) == result

    if remove_server_attrs:
        submgr.remove_server(remove_server_attrs)

        # The server_id should be missing and we get ValueError
        try:
            submgr.get_all_subscriptions(remove_server_attrs)
            assert False
        except ValueError:
            pass

    # Test for final counts after any remove test definitions
    if 'final_listenercount' in exp_result:
        assert len(submgr.get_all_destinations(server_id)) == \
            exp_result['final_listener_count']

    if 'final_filter_count' in exp_result:
        assert len(submgr.get_all_filters(server_id)) == \
            exp_result['final_filter_count']

    if 'final_subscription_count' in exp_result:
        assert len(submgr.get_all_subscriptions(server_id)) == \
            exp_result['final_subscription_count']


TESTCASES_SUBMGR_MODIFY = [

    # Testcases for Subscription Tests multiple valid and invalid options
    # in __init__, add_server, add_filter, add_destinations

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * submgr_id: submgr id
    #   * connection_attrs: FakedWBEMConnection attributes. this variable causes
    #     an add_server to be executed. If it is a list, multiple
    #     a server is added for each dict in the list. The attributes of
    #     the FakedWBEMConnection can be set as attributes. Note: this test
    #     already sets the interop namespace automatically
    #   * filter_attrs: dictionary of attributes for add_filter. If it is
    #     a list, one filter is added for each dict in the list
    #   * dest_attrs: dictionary of attributes for add_destination().
    #   * modify_dest: dictionary of attributes for modify_instance of listener
    #     destination.
    #   * subscription_attrs: dictionary or list dictionaries where each
    #     dictionary is a dictionary that defines either the attributes of
    #     the subscription or to make it simpler to define tests, an integer
    #     for the position in the appropriate list for the required
    #     filter or destination.
    #   * modify_filter_attrs: defines inst/properties properties to modify
    #     where instances are modified on the server external to
    #     SubscriptionManager (i.e. directly with conn.ModifyInstance)
    #     It can be either a pair of attributes or a CIMProperty
    #     If it is a pair of attributes the attributes are:
    #       1. The index to the filter_dest to modify
    #       2. dict of properties to modify
    #   * modify_listener_dest_attrs: defines inst/properties properties to
    #     modify. The attributes are:
    #       1. The index to the listener_dest to modify
    #       2. dict of properties to modify
    #   * modify_subscription_attrs: defines inst/properties properties to
    #     modify. The attributes are:
    #       1. The index to the subscription_dest to modify
    #       2. dict of properties to modify
    #   * exp_result: Expected dictionary items, for validation. This includes:
    #       * server_id - The expected server_id
    #       * listener_count - Count of expected destinations
    #       * filter_count - Count of expected filters.
    #       * dest_props: TODO: Will get doc from later pr
    #       * TODO: Add validation for filter, subscriptions
    #       * subscription_count - Count of expected subscriptions.
    #
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Modify fails, No modify property returns not supported from provider",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=None,
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            subscription_attrs=None,
            modify_filter_attrs=None,
            modify_dest_attrs=[0, []],
            modify_subscription_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1)
        ),
        CIMError, None, OK
    ),
    (
        "Modify OK. Property modified",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            subscription_attrs=None,
            modify_dest_attrs=None,
            modify_filter_attrs=[0, [('IndividualSubscriptionSupported',
                                      True)]],
            modify_subscription_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_props=[0,
                                          [('IndividualSubscriptionSupported',
                                            True)]]),
        ),
        None, None, OK
    ),
    (
        "Modify fails, Subscription exists",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=True),
            modify_dest_attrs=None,
            modify_filter_attrs=[0, [('IndividualSubscriptionSupported',
                                      True)]],
            modify_subscription_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_props=[0,
                                          [('IndividualSubscriptionSupported',
                                            True)]]),
        ),
        CIMError, None, OK
    ),
    (
        "Modify fails, Filter query language, mock provider forbids change",
        dict(
            submgr_id="ValidID",
            connection_attrs=dict(url=None),
            filter_attrs=dict(server_id="http://FakedUrl:5988",
                              source_namespaces='root/interop',
                              query="SELECT * from blah",
                              query_language='WQL',
                              owned=True,
                              filter_id="Filter1", name=None),
            dest_attrs=dict(server_id="http://FakedUrl:5988",
                            listener_url="http://localhost:5000",
                            owned=True,
                            destination_id='id1'),
            subscription_attrs=dict(server_id="http://FakedUrl:5988",
                                    filter_path=0,
                                    destination_paths=0,
                                    owned=True),
            modify_dest_attrs=None,
            modify_filter_attrs=[0, [('QueryLanguage',
                                      'WQL')]],
            modify_subscription_attrs=None,
            exp_result=dict(server_id="http://FakedUrl:5988",
                            listener_count=1,
                            filter_props=[0,
                                          [('QueryLanguage',
                                            'WQL')]]),
        ),
        CIMError, None, OK
    ),

    # TODO: The following tests are not in the test list.
    # 1. Test for missing required_property. This one very difficult because
    #    list of required properties is also properties always set by
    #    SubscriptionManager
    # 2. Invalid Namespace on __init__ of mock subscription providers
    # 3. __repr__ for each mock subscription providers
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_SUBMGR_MODIFY)
@simplified_test_function
def test_submgr_modify(testcase, submgr_id, connection_attrs,
                       filter_attrs, dest_attrs, subscription_attrs,
                       modify_filter_attrs, modify_dest_attrs,
                       modify_subscription_attrs, exp_result):
    # pylint: disable=unused-argument
    """
    Tests the ability to modify instances of the subscription classes in the
    pywbem_mock subscription providers. This creates filters, listener
    destinations, and subscriptions and then tests with ModifyInstance.
    Note: The subscription manager does not include ModifyInstance methods
    for the indication subscription classes.
    """
    skip_if_moftab_regenerated()

    def define_connection(connection_attrs):
        """
        Create a new mock connection with the attributes defined in
        connection_attrs.
        This method also does an add_server with the created mock server.
        Returns:
            returns the server_id associated with the connection and the
            connection object.
        """
        mock_server = WbemServerMock(
            interop_ns='interop',
            server_mock_data=SUBSCRIPTION_WBEM_SERVER_MOCK_DICT,
            **connection_attrs)
        conn = mock_server.wbem_server.conn
        return (submgr.add_server(mock_server.wbem_server), conn)

    # The code to be tested starts here but encompasses most of this function
    # depending on the function parameters  We did not set a specific place
    # above which there should never be an exception. Depending on the tests,
    # exceptions could occur throughout the code below.
    submgr = WBEMSubscriptionManager(submgr_id)

    # Test result arrays for one or more server_ids. Each of the add and
    # remove methods will insert the results into these lists for validation
    # of the returns from these methods.
    server_id = None
    server_ids = []
    server_conn = None
    added_filters = []
    added_destinations = []
    added_subscriptions = []

    # The following code adds connections and filters
    # If connection_attrs exists, set up the server and save the server_ids
    # for later use in the test
    if connection_attrs:
        if isinstance(connection_attrs, list):
            # DO not allow multiple connections for this test.
            assert False
            for connection_attr in connection_attrs:
                server_ids.append(define_connection(connection_attr))
        else:
            server_id, server_conn = define_connection(connection_attrs)

    # Test for add filter, adds one or more filters
    if filter_attrs:
        if isinstance(filter_attrs, dict):
            filter_attrs = [filter_attrs]
        for attr in filter_attrs:
            added_filters.append(submgr.add_filter(**attr))

    # Test for add destinations.

    if dest_attrs:
        if isinstance(dest_attrs, dict):
            dest_attrs = [dest_attrs]
        for attr in dest_attrs:
            added_destinations.append(submgr.add_destination(**attr))

    # Test for adding subscriptions based on the attributes in the dest
    # definition. Here, the filter and dest components are integers defining
    # The entry in the already created object lists to use for the path
    # in the subscription.
    if subscription_attrs:
        if isinstance(subscription_attrs, dict):
            subscription_attrs = [subscription_attrs]
        for attr in subscription_attrs:
            if isinstance(attr['filter_path'], int):
                attr['filter_path'] = added_filters[attr['filter_path']].path
            if isinstance(attr['destination_paths'], int):
                attr['destination_paths'] = \
                    added_destinations[attr['destination_paths']].path
            elif isinstance(attr['destination_paths'], list):
                dest_paths = []
                for item in attr['destination_paths']:
                    dest_paths.append(
                        added_destinations[item].path)
                attr['destination_paths'] = dest_paths
            rslts = submgr.add_subscriptions(**attr)
            added_subscriptions.extend(rslts)

    if modify_dest_attrs:
        instance_to_modify = added_destinations[modify_dest_attrs[0]]
        modified_instance = instance_to_modify.copy()
        modified_instance.update(modify_dest_attrs[1])
        pl = [p[0] for p in modify_dest_attrs[1]]
        server_conn.ModifyInstance(modified_instance, PropertyList=pl)

    if modify_filter_attrs:
        instance_to_modify = added_filters[modify_filter_attrs[0]]
        modified_instance = instance_to_modify.copy()
        modified_instance.update(modify_filter_attrs[1])
        pl = [p[0] for p in modify_filter_attrs[1]]
        server_conn.ModifyInstance(modified_instance, PropertyList=pl)

    if modify_subscription_attrs:
        instance_to_modify = added_subscriptions[modify_subscription_attrs[0]]
        modified_instance = instance_to_modify.copy()
        modified_instance.update(modify_subscription_attrs[1])
        pl = [p[0] for p in modify_subscription_attrs[1]]
        server_conn.ModifyInstance(modified_instance)

    if 'dest_props' in exp_result:
        dest_props = exp_result['dest_props']
        orig_instance = added_destinations[dest_props[0]]
        updated_instance = server_conn.GetInstance(orig_instance.path)
        props = dest_props[1]
        for prop, value in props:
            assert updated_instance[prop] == value

    # test for listener_destinations created
    if 'listener_count' in exp_result:
        all_dests = submgr.get_all_destinations(server_id)
        assert len(all_dests) == exp_result['listener_count']

        owned_dests = submgr.get_owned_destinations(server_id)

        dest_prefix = 'pywbemdestination:{0}:'.format(submgr_id,)
        owned = [d for d in all_dests if d['Name'].startswith(dest_prefix)]
        assert set(owned) == set(owned_dests)

    if 'filter_count' in exp_result:
        all_filters = submgr.get_all_filters(server_id)
        assert len(all_filters) == exp_result['filter_count']

    if 'subscription_count' in exp_result:
        all_subs = submgr.get_all_subscriptions(server_id)
        assert len(all_subs) == exp_result['subscription_count']



"""
End2end tests of operation timeout from a test class in the
OpenPegasus WBEM server.

"""

from __future__ import absolute_import, print_function

import re

import datetime

from .utils.pytest_extensions import skip_if_unsupported_capability

# Note: The wbem_connection fixture uses the es_server fixture, and
# due to the way py.test searches for fixtures, it also must be imported.
# pylint: disable=unused-import,wrong-import-order, relative-beyond-top-level
from .utils.pytest_extensions import wbem_connection  # noqa: F401
from pytest_easy_server import es_server  # noqa: F401
from .utils.pytest_extensions import default_namespace  # noqa: F401
# pylint: enable=unused-import,wrong-import-order, relative-beyond-top-level

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
# pylint: disable=relative-beyond-top-level, disable=redefined-builtin
from ..utils import import_installed

pywbem = import_installed('pywbem')
from pywbem import Uint32, CIMError, ConnectionError, CIMClassName, \
    TimeoutError  # noqa: E402 pylint: disable=redefined-builtin
# pylint: enable=relative-beyond-top-level, enable=redefined-buitin
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# Identity of the OpenPegasus namespace, class, and method that implements
# the delayed response. NOTE: This is only available on OpenPegasus version
# 2.15+ available from github.
NAMESPACE = 'test/testprovider'
TESTCLASS = 'Test_CLITestProviderClass'
TESTMETHOD = 'delayedMethodResponse'


class ElapsedTimer(object):
    """
        Set up elapsed time timer. Calculates time between initiation
        and access.
    """

    def __init__(self):
        """ Initiate the object with current time"""
        self.start_time = datetime.datetime.now()

    def reset(self):
        """ Reset the start time for the timer"""
        self.start_time = datetime.datetime.now()

    def elapsed_ms(self):
        """ Get the elapsed time in milliseconds. returns floating
            point representation of elapsed time in seconds.
        """
        dt = datetime.datetime.now() - self.start_time
        return ((dt.days * 24 * 3600) + dt.seconds) * 1000 + \
            dt.microseconds / 1000.0

    def elapsed_sec(self):
        """ get the elapsed time in seconds. Returns floating
            point representation of time in seconds
        """
        return self.elapsed_ms() / 1000


def run_timeout_test(conn, timeout, response_delay, timeout_exp):
    """
    Test a single timeout, expected response_delay and timeout_exp boolean
    Tests for correct response type (good response if the delay is
    less than the timeout and timeout exception if the delay is greater
    than the timeout.
    """
    opttimer = ElapsedTimer()
    conn.timeout = timeout
    try:
        # Confirm server working with a simple request.
        conn.GetClass('CIM_ManagedElement')
        assert conn.timeout == timeout
        class_path = CIMClassName(TESTCLASS, namespace=NAMESPACE)
        result = conn.InvokeMethod(TESTMETHOD, class_path,
                                   [('delayInSeconds', Uint32(response_delay))])

        assert result[0] == Uint32(0)
        assert result[1]['delayInSeconds'] == response_delay

        execution_time = opttimer.elapsed_sec()

        assert not timeout_exp, 'Timeout expected but not rcvd., Fail: ' \
            'url={}, conn.timeout={}, timeout={}, response delay exp={}, '  \
            'execution time={}'.format(conn.url, conn.timeout, timeout,
                                       response_delay, execution_time)

    # ConnectionError exception: If this does not include in the response
    # message the text string, "Read timed out" the test fails.
    # If it does include that string the timeout is processed. See issue # 3075
    # Timeout returns this message text:
    # "HTTPConnectionPool(host='localhost', port=15988): Read timed out. "
    # #(read timeout=4)"
    except ConnectionError as ce:
        execution_time = opttimer.elapsed_sec()

        if re.search(r'^HTTPS?ConnectionPool(.*)Read timed out. (.*)$',
                     ce.args[0]):
            time_diff = abs(execution_time - timeout)
            assert timeout_exp, 'Timeout not expected: url={}, ' \
                'conn.timeout={}, ' \
                'timeout={}, response_delay_exp={}, execution time={}'. \
                format(conn.url, conn.timeout, timeout, response_delay,
                       execution_time)
            assert time_diff <= 1, 'Timeout Time - response delay diff to ' \
                'great, Failed: url={}, timeout={}, response delay exp={}, ' \
                'execution time={} exp - act diff={}, exception={}'. \
                format(conn.url, timeout, response_delay, execution_time,
                       time_diff, ce)

        else:
            assert False, 'ConnectionError Fail: url={}, conn.timeout={}, '  \
                'timeout={}, response_delay={}, execution time={}, '  \
                'exception={}'. \
                format(conn.url, conn.timeout, timeout, response_delay,
                       execution_time, ce)

    # This exception tests against expected result. It terminates the
    # test if the timeout is not expected or time difference between
    # response delay and timeout is too great.
    except TimeoutError as te:
        execution_time = opttimer.elapsed_sec()
        time_diff = abs(execution_time - timeout)
        assert timeout_exp, 'Timeout not expected: url={}, ' \
            'conn.timeout={}, ' \
            'timeout={}, response_delay_exp={}, execution time={}'. \
            format(conn.url, conn.timeout, timeout, response_delay,
                   execution_time)
        assert time_diff <= 1, 'Timeout Time - response delay diff to great, ' \
            'Failed: url={}, timeout={}, response delay exp={}, ' \
            'execution time={} exp - act diff={}, exception={}'. \
            format(conn.url, timeout, response_delay, execution_time,
                   time_diff, te)

    # Any CIMError
    except CIMError as ce:
        execution_time = opttimer.elapsed_sec()
        time_diff = abs(execution_time - timeout)
        assert not timeout_exp, 'CIMError exception: url={}, ' \
            'conn.timeout={}, timeout={}, response_delay_exp={}, ' \
            'execution time={} exception={}'. \
            format(conn.url, conn.timeout, timeout, response_delay,
                   execution_time, ce)


def test_timeouts(wbem_connection):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test indication subscription to a container and reception of indications
    from the container.
    """
    skip_if_unsupported_capability(wbem_connection, 'interop')

    # Table of tests. Each entry is:
    #    timeout (sec), response_delay(sec), bool(time_out_expected flag)
    timeout_tests = [{'timeout': 8,
                      'response_delay': 4,
                      'timeout_exp': False},
                     {'timeout': 4,
                      'response_delay': 8,
                      'timeout_exp': True}, ]
    for kwargs in timeout_tests:
        run_timeout_test(wbem_connection, **kwargs)
